"""
Provides a simple framework for unifying different parallelism approaches.
Currently primarily targets multiprocessing and mpi4py, but also should work
with Ray. Dask will require more work unfortunately...
"""

import abc, functools, multiprocessing as mp, multiprocessing.pool as mp_pool, typing, uuid, os
import contextlib
import enum

import numpy as np, pickle, time

from .. import Devutils as dev
from ..Scaffolding import Logger, NullLogger, ObjectRegistry
from .SharedMemory import SharedObjectManager, SharedMemoryList, SharedMemoryDict

__all__ = [
    "Parallelizer",
    "MultiprocessingParallelizer",
    "MPIParallelizer",
    "SerialNonParallelizer",
    "SendRecieveParallelizer"
]

class CallerContract:
    """
    Provides a structure so that a main process and child process can
    be synchronized in their MPI-like calls
    """

    def __init__(self, calls):
        self.calls = calls
        self.which_call = 0

    def handle_call(self, caller, next_call):
        if next_call != self.calls[self.which_call]:
            raise ValueError("caller {} tried to call {} but expected to call {}".format(
                caller,
                next_call,
                self.calls[self.which_call]
            ))
        self.which_call = (self.which_call + 1) % len(self.calls)

class ChildProcessRuntimeError(RuntimeError):
    ...

class Parallelizer(metaclass=abc.ABCMeta):
    """
    Abstract base class to help manage parallelism.
    Provides the basic API that all parallelizers can be expected
    to conform to.
    Provides effectively the union of operations supported by
    `mp.Pool` and `MPI`.
    There is also the ability to lookup and register 'named'
    parallelizers, since we expect a single program to not
    really use more than one.
    This falls back gracefully to the serial case.
    """

    ####################################################################
    #  CONTEXT MANAGEMENT:
    #   details for initialization/registration/getting and setting
    #   up a parallelizer
    #
    _par_registry = None
    default_printer = print
    base_log_level = Logger.LogLevel.MoreDebug
    def __init__(self, logger=None, contract=None, uid=None,
                 initialization_function=None,
                 initialization_args=None,
                 initialization_kwargs=None
                 ):
        self._active_sentinel=0
        self._pickle_prot = None
        if logger is None:
            logger = Logger()
        self.logger = logger
        self.contract = (
            contract
                if contract is None or isinstance(contract, CallerContract) else
            CallerContract(contract)
        )
        self._default_stack = None
        self.uid = uuid.uuid1()
        self._pid = None
        self.initialization_args = () if initialization_args is None else initialization_args
        self.initialization_kwargs = {} if initialization_kwargs is None else initialization_kwargs
        self.initialization_function = (
            functools.partial(initialization_function,
                              *initialization_args,
                              **initialization_kwargs)
                if initialization_function is not None else
            initialization_function
        )
        # if printer is None:
        #     self._logger = Logger()
        #     self._default_printer = self._logger.log_print
        # else:
        #

    @classmethod
    def load_registry(cls):
        if cls._par_registry is None:
            cls._par_registry = ObjectRegistry(default=SerialNonParallelizer())
        return cls._par_registry
    @property
    def parallelizer_registry(self):
        return self.load_registry()

    @classmethod
    def get_default(cls):
        """
        For compat.

        :return:
        :rtype:
        """
        return cls.lookup(None, construct=False)
    class Backends(enum.Enum):
        Serial = "serial"
        MPI = "mpi"
        Multiprocessing = "multiprocessing"
    @classmethod
    def get_paralellizer_types(cls):
        return {
            cls.Backends.Serial:SerialNonParallelizer,
            cls.Backends.MPI:MPIParallelizer,
            cls.Backends.Multiprocessing:MultiprocessingParallelizer,
        }
    parallelizer_dispatch = None
    @classmethod
    def lookup(cls, key, construct=True) -> 'Parallelizer':
        """
        Checks in the registry to see if a given parallelizer is there
        otherwise returns a `SerialNonParallelizer`.
        :param key:
        :type key:
        :return:
        :rtype:
        """
        if isinstance(key, Parallelizer):
            return key
        else:
            if key is None or isinstance(key, str):
                new = cls.load_registry().lookup(key)
            else:
                new = None
            if construct and (
                    new is None
                    or isinstance(new, SerialNonParallelizer)
            ) and dev.is_option_spec_like(key):
                cls.parallelizer_dispatch = dev.handle_uninitialized(
                    cls.parallelizer_dispatch,
                    lambda: dev.OptionsMethodDispatch(
                        cls.get_paralellizer_types,
                        methods_enum=cls.Backends,
                        ignore_bad_enum_keys=True
                    )
                )
                type, opts = cls.parallelizer_dispatch.resolve(
                    key
                )
                if type is not None:
                    new = type(**opts)
            return new

    def register(self, key):
        """
        Checks in the registry to see if a given parallelizer is there
        otherwise returns a `SerialNonParallelizer`.
        :param key:
        :type key:
        :return:
        :rtype:
        """
        self.parallelizer_registry.register(key, self)

    @property
    def active(self):
        return self._active_sentinel > 0

    def set_initializer(self, func, *args, **kwargs):
        self.initialization_function = functools.partial(func, *args, **kwargs)
        self.run(self.initialization_function)
        # self.pool.map(
        #     self._initialize_worker,
        #     [self.initialization_function] * (self.nproc)
        # )
        # self.pool.initializer = functools.partial(
        #     self._initialize_worker,
        #     self.initialization_function
        # )
    @abc.abstractmethod
    def initialize(self):
        """
        Initializes a parallelizer
        if necessary
        :return:
        :rtype:
        """
        raise NotImplementedError("Parallelizer is an abstract base class")
    @abc.abstractmethod
    def finalize(self, exc_type, exc_val, exc_tb):
        """
        Finalizes a parallelizer (if necessary)
        if necessary
        :return:
        :rtype:
        """
        raise NotImplementedError("Parallelizer is an abstract base class")

    def __enter__(self):
        """
        Allows the parallelizer context to be set
        using `with`
        :return:
        :rtype:
        """
        if not self.active:
            self._default_stack = self.load_registry().temp_default(self)
            self._default_stack.__enter__()
            self.register(self.uid)
            self.initialize()
            self._pickle_prot = pickle.DEFAULT_PROTOCOL
            pickle.DEFAULT_PROTOCOL = pickle.HIGHEST_PROTOCOL
        self._active_sentinel += 1
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Allows the parallelizer context to be unset
        using `with`
        :param exc_type:
        :type exc_type:
        :param exc_val:
        :type exc_val:
        :param exc_tb:
        :type exc_tb:
        :return:
        :rtype:
        """
        self._active_sentinel -= 1
        if not self.active:
            if self._default_stack is not None:
                self._default_stack.__exit__(exc_type, exc_val, exc_tb)
                self._default_stack = None
            self.finalize(exc_type, exc_val, exc_tb)
            # self._pickle_prot = pickle.DEFAULT_PROTOCOL
            pickle.DEFAULT_PROTOCOL = self._pickle_prot

    ####################################################################
    #  CONTROL FLOW:
    #   clean ways to manage flow of program
    #
    class InMainProcess:
        """
        Singleton representing being on the
        main process
        """
    class InWorkerProcess:
        """
        Singleton representing being on a
        worker process
        """
    @property
    @abc.abstractmethod
    def on_main(self):
        """
        Returns whether or not the executing process is the main
        process or not
        :return:
        :rtype:
        """

    @staticmethod
    def main_restricted(func):
        """
        A decorator to indicate that a function should only be
        run when on the main process
        :param func:
        :type func:
        :return:
        :rtype:
        """

        def main_process_func(*args, parallelizer=None, **kwargs):
            if parallelizer is None:
                parallelizer = Parallelizer.lookup(None)
            if parallelizer.on_main:
                kwargs['parallelizer'] = parallelizer
                return func(*args, **kwargs)
            else:
                return Parallelizer.InWorkerProcess

        return main_process_func

    @staticmethod
    def worker_restricted(func):
        """
        A decorator to indicate that a function should only be
        run when on a worker process
        :param func:
        :type func:
        :return:
        :rtype:
        """

        @functools.wraps(func)
        def worker_process_func(*args, parallelizer=None, **kwargs):
            if parallelizer is None:
                parallelizer = Parallelizer.lookup(None)
            if not parallelizer.on_main:
                kwargs['parallelizer'] = parallelizer
                return func(*args, **kwargs)
            else:
                return Parallelizer.InMainProcess

        return worker_process_func

    ####################################################################
    #  MPI API:
    #   Require core methods to be implemented to support the
    #   high-level MPI api.
    #   Currently only blocking operations are expected to be supported,
    #   since it's unlikely we'll ever have a use case for async calls
    #
    @abc.abstractmethod
    def send(self, data, loc, **kwargs):
        """
        Sends data to the process specified by loc

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        raise NotImplementedError("Parallelizer is an abstract base class")
    @abc.abstractmethod
    def receive(self, data, loc, **kwargs):
        """
        Receives data from the process specified by loc

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        raise NotImplementedError("Parallelizer is an abstract base class")
    @abc.abstractmethod
    def broadcast(self, data, **kwargs):
        """
        Sends the same data to all processes

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        raise NotImplementedError("Parallelizer is an abstract base class")
    @abc.abstractmethod
    def scatter(self, data, locs=None, return_locs=False, **kwargs):
        """
        Performs a scatter of data to the different
        available parallelizer processes.
        *NOTE:* unlike in the MPI case, `data` does not
        need to be evenly divisible by the number of available
        processes

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        raise NotImplementedError("Parallelizer is an abstract base class")
    @abc.abstractmethod
    def gather(self, data, **kwargs):
        """
        Performs a gather of data from the different
        available parallelizer processes

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        raise NotImplementedError("Parallelizer is an abstract base class")

    ####################################################################
    #  POOL API:
    #   require core methods to be implemented to act like an `mp.Pool`
    #
    @abc.abstractmethod
    def map(self, function, data, extra_args=None, extra_kwargs=None, **kwargs):
        """
        Performs a parallel map of function over
        the held data on different processes

        :param function:
        :type function:
        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        raise NotImplementedError("Parallelizer is an abstract base class")

    @abc.abstractmethod
    def starmap(self, function, data, extra_args=None, extra_kwargs=None, **kwargs):
        """
        Performs a parallel map with unpacking of function over
        the held data on different processes

        :param function:
        :type function:
        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        raise NotImplementedError("Parallelizer is an abstract base class")

    @abc.abstractmethod
    def apply(self, func, *args, main_kwargs=None, cleanup=True, **kwargs):
        """
        Runs the callable `func` in parallel
        :param func:
        :type func:
        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        raise NotImplementedError("Parallelizer is an abstract base class")

    def run(self, func, *args, comm=None, main_kwargs=None, **kwargs):
        """
        Calls `apply`, but makes sure state is handled cleanly

        :param func:
        :type func:
        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        with self:
            return self.apply(func, *args, comm=comm, main_kwargs=main_kwargs, **kwargs)

    mode_map = {}
    @classmethod
    def from_config(cls,
                    mode=None,
                    **kwargs
                    ):
        if mode not in cls.mode_map:
            raise KeyError("don't know what to do with parallelization mode '{}'".format(
                mode
            ))
        par_cls = cls.mode_map[mode]
        return par_cls.from_config(**kwargs)

    ####################################################################
    #  UTILITY API:
    #   necessary things to make API more useable
    #
    @property
    def nprocs(self):
        """
        Returns the number of processes the parallelizer has
        to work with
        :return:
        :rtype:
        """
        return self.get_nprocs()
    @abc.abstractmethod
    def get_nprocs(self):
        """
        Returns the number of processes
        :return:
        :rtype:
        """
        raise NotImplementedError("Parallelizer is an abstract base class")
    @property
    def id(self):
        """
        Returns some form of identifier for the current process
        :return:
        :rtype:
        """
        return self.get_id()
    @property
    def pid(self):
        if self._pid is None:
            self._pid = os.getpid()
        return self._pid
    @abc.abstractmethod
    def get_id(self):
        """
        Returns the id for the current process
        :return:
        :rtype:
        """
        raise NotImplementedError("Parallelizer is an abstract base class")

    @property
    def printer(self):
        if self.logger is None:
            return self.default_printer
        else:
            return self.logger.log_print
    @printer.setter
    def printer(self, p):
        self._printer = p
    def main_print(self, *args, **kwargs):
        """
        Prints from the main process
        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        self.printer(*args, **kwargs)
    def worker_print(self, *args, **kwargs):
        """
        Prints from a main worker process
        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        self.printer(" ".join(["On Worker {} ({}):".format(self.id, self.pid), *(str(x) for x in args)]), **kwargs)
    def print(self, *args, where='both', **kwargs):
        """
        An implementation of print that operates differently on workers than on main
        processes

        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        if self.on_main:
            if where in {'both', 'main'}:
                return self.main_print(*args, **kwargs)
        else:
            if where in {'both', 'worker'}:
                return self.worker_print(*args, **kwargs)

    @abc.abstractmethod
    def wait(self):
        """
        Causes all processes to wait until they've met up at this point.
        :return:
        :rtype:
        """
        raise NotImplementedError("Parallelizer is an abstract base class")

    def __repr__(self):
        try:
            id = self.id
        except:
            id = None
        try:
            nprocs = self.nprocs
        except:
            nprocs = None
        return "{}(id={}, nprocs={}, uuid={})".format(type(self).__name__, id, nprocs, self.uid)

    def share(self, obj):
        """
        Converts `obj` into a form that can be cleanly used with shared memory via a `SharedObjectManager`

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        if isinstance(obj, dict):
            sharer = SharedMemoryDict(obj, parallelizer=self)
        elif isinstance(obj, (list, tuple)):
            sharer = SharedMemoryList(obj, parallelizer=self)
        else:
            sharer = SharedObjectManager(obj, parallelizer=self)
            sharer.share()

        return sharer

class SendRecieveParallelizer(Parallelizer):
    """
    Parallelizer that implements `scatter`, `gather`, `broadcast`, and `map`
    based on just having a communicator that supports `send` and `receive methods
    """

    class ReceivedError:
        def __init__(self, error):
            self.error = error
    class SendReceieveCommunicator(metaclass=abc.ABCMeta):
        """
        A base class that provides an interface for
        sending/receiving data from a specific subprocesses
        """
        @property
        @abc.abstractmethod
        def locations(self):
            """
            Returns the list of locations known by the
            `SendReceieverCommunicator`
            :return:
            :rtype:
            """
        @property
        @abc.abstractmethod
        def location(self):
            """
            Returns the _current_ location
            :return:
            :rtype:
            """
        @abc.abstractmethod
        def send(self, data, loc, **kwargs):
            """
            Sends the specified data to loc
            :param data:
            :type data:
            :param loc:
            :type loc:
            :param kwargs:
            :type kwargs:
            :return:
            :rtype:
            """
        @abc.abstractmethod
        def receive(self, data, loc, **kwargs):
            """
            Receives the specified data from loc
            :param data:
            :type data:
            :param loc:
            :type loc:
            :param kwargs:
            :type kwargs:
            :return:
            :rtype:
            """
    @property
    @abc.abstractmethod
    def comm(self):
        """
        Returns the communicator used by the paralellizer
        :return:
        :rtype: SendRecieveParallelizer.SendReceieveCommunicator
        """
    def send(self, data, loc, **kwargs):
        """
        Sends data to the process specified by loc

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        if self.contract is not None:
            self.contract.handle_call(self, "send")
        return self.comm.send(data, loc, **kwargs)
    def receive(self, data, loc, **kwargs):
        """
        Receives data from the process specified by loc

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        if self.contract is not None:
            self.contract.handle_call(self, "receive")
        val = self.comm.receive(data, loc, **kwargs)
        if isinstance(val, self.ReceivedError):
            raise val.error
        else:
            return val
    def broadcast(self, data, locs=None, **kwargs):
        """
        Sends the same data to all processes

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        if self.contract is not None:
            self.contract.handle_call(self, "broadcast")
        if self.on_main:
            if locs is None:
                locs = list(self.comm.locations)[1:]
            for i in locs:
                self.comm.send(data, i, **kwargs)
            return data
        else:
            return self.comm.send(data, self.comm.location) # effectively a receive...
    def scatter(self, data, locs=None, return_locs=False, **kwargs):
        """
        Performs a scatter of data to the different
        available parallelizer processes.
        *NOTE:* unlike in the MPI case, `data` does not
        need to be evenly divisible by the number of available
        processes

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        if self.contract is not None:
            self.contract.handle_call(self, "scatter")
        if self.on_main:
            if locs is None:
                locs = list(self.comm.locations)[1:]
            locs = list(locs) # gotta be safe
            nlocs = len(locs) + 1
            chunk_size = len(data) // nlocs
            chunk_sizes = [chunk_size] * nlocs
            chunk_coverage = (chunk_size*nlocs)
            for i in range(len(data)-chunk_coverage):
                chunk_sizes[i] += 1
            s = chunk_sizes[0]
            main_data = data[:s]
            for i, b in zip(locs, chunk_sizes[1:]):
                chunk = data[s:s+b]
                self.comm.send(chunk, i, **kwargs)
                s+=b
            if not return_locs:
                return main_data
            else:
                return main_data, locs[:len(chunk_sizes)-1]
        else:
            res = self.comm.send(data, self.comm.location)
            if return_locs:
                return res, None
            else:
                return res # effectively a receive...
    def gather(self, data, locs=None, **kwargs):
        """
        Performs a gather of data from the different
        available parallelizer processes

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        if self.contract is not None:
            self.contract.handle_call(self, "gather")
        if self.on_main:
            if locs is None:
                locs = list(self.comm.locations)[1:] # gotta be safe
            else:
                locs = list(locs)
            nlocs = len(locs) # we reserve space for the main thread
            recv = [None] * (nlocs + 1)
            recv[0] = data
            all_numpy = True
            for n, i in enumerate(locs):
                res = self.comm.receive(data, i, **kwargs)
                if not isinstance(res, np.ndarray):
                    all_numpy=False
                if isinstance(res, Exception):
                    raise res
                recv[n+1] = res
            # if all_numpy:
            #     # special case
            #     try:
            #         recv = np.concatenate(recv, axis=0)
            #     except:
            #         pass
            return recv
        else:
            return self.comm.receive(data, self.comm.location, **kwargs) # effectively a send...
    def map(self, func, data, extra_args=None, extra_kwargs=None,
            vectorized=False,
            aggregate=True,
            **kwargs):
        """
        Performs a parallel map of function over
        the held data on different processes

        :param function:
        :type function:
        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        # self.wait()
        self.print("Scattering Data", log_level=Parallelizer.base_log_level)
        data, locs = self.scatter(data, return_locs=True, **kwargs)
        # self.wait()
        self.print("Broadcasting Extra Args", log_level=Parallelizer.base_log_level)
        extra_args = self.broadcast(extra_args, **kwargs)
        # self.wait()
        self.print("Broadcasting Extra Kwargs", log_level=Parallelizer.base_log_level)
        extra_kwargs = self.broadcast(extra_kwargs, **kwargs)
        if extra_args is None:
            extra_args = ()
        if extra_kwargs is None:
            extra_kwargs = {}
        # try:
        if vectorized:
            evals = func(data, *extra_args, **extra_kwargs)
        else:
            evals = [
                func(sub_data, *extra_args, **extra_kwargs)
                for sub_data in data
            ]
        # except Exception as e:
        #     self.gather(e, **kwargs)
        #     raise
        res = self.gather(evals, locs=locs, **kwargs)
        if self.on_main:
            if aggregate:
                if len(res) == 1:
                    res = res[0]
                else:
                    _ = []
                    for r in res:
                        _.extend(r)
                    res = _
            return res
        else:
            return Parallelizer.InWorkerProcess
    def starmap(self, func, data, extra_args=None, extra_kwargs=None, **kwargs):
        """
        Performs a parallel map with unpacking of function over
        the held data on different processes

        :param function:
        :type function:
        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        data = self.scatter(data, **kwargs)
        extra_args = self.broadcast(extra_args)
        extra_kwargs = self.broadcast(extra_kwargs)
        if extra_args is None:
            extra_args = ()
        if extra_kwargs is None:
            extra_kwargs = {}
        # try:
        evals = [func(*sub_data, *extra_args, **extra_kwargs) for sub_data in data]
        # except Exception as e:
        #     self.gather(e, **kwargs)
        #     raise
        res = self.gather(evals, **kwargs)
        if self.on_main:
            return res
            # return sum(res, [])
        else:
            return Parallelizer.InWorkerProcess

    def wait(self):
        """
        Causes all processes to wait until they've met up at this point.
        :return:
        :rtype:
        """

        # flag = self.broadcast("SyncFlag")
        self.gather("SyncFlag") # sends data to main
        self.broadcast("SyncFlag")

class MultiprocessingParallelizer(SendRecieveParallelizer):
    """
    Parallelizes using a  process pool and a runner
    function that represents a "main loop".
    """

    class SendRecvQueuePair:
        def __init__(self, id:int, manager:'mp.managers.SyncManager'):
            self.id=id
            self.send_queue = manager.Queue()
            self.receive_queue = manager.Queue()
            self.init_flag = manager.Event()
        def __repr__(self):
            return "{}({})".format(type(self).__name__, self.id)
    class PoolCommunicator(SendRecieveParallelizer.SendReceieveCommunicator):
        """
        Defines a serializable process communicator
        that allows communication with a managed `mp.Pool`
        to support `send` and `receive` and therefore to
        support the rest of the necessary bits of the MPI API
        """
        def __init__(self,
                     parent: 'MultiprocessingParallelizer',
                     id:int,
                     queues: typing.Iterable['MultiprocessingParallelizer.SendRecvQueuePair'],
                     initialization_timeout:float=None,
                     group:typing.Iterable['MultiprocessingParallelizer.PoolCommunicator']=None
                     ):
            self.parent = parent
            self.id = id
            self.queues = tuple(queues)
            self.initialization_timeout=initialization_timeout
            self.group = tuple(group) if group is not None else None

        def __repr__(self):
            return "{}({})".format(
                type(self).__name__,
                self.id
            )

        class PoolError(Exception):
            """
            For errors arising from Pool operations
            """
            pass

        def initialize(self):
            """
            Performs initialization of the communicator
            (basically just waits until all threads say all is well)
            :return:
            :rtype:
            """
            if not self.queues[self.id].init_flag.is_set():
                self.parent.print("setting init flag on {id}", id=self.id, log_level=Parallelizer.base_log_level)
                self.queues[self.id].init_flag.set()
            if self.parent.on_main:
                for i, q in enumerate(self.queues):
                    if not q.init_flag.is_set():
                        self.parent.print("checking init flag on {i}".format(i=i), log_level=Parallelizer.base_log_level)
                        wat = q.init_flag.wait(self.initialization_timeout)
                        if not wat:
                            raise self.PoolError("Failed to initialize pool")

        def reset(self):
            """
            Performs initialization of the communicator
            (basically just waits until all threads say all is well)
            :return:
            :rtype:
            """
            for i, q in enumerate(self.queues):
                q.init_flag.clear()
            # if not self.queues[self.id].init_flag.is_set():
            #     self.parent.print("setting init flag on {id}", id=self.id, log_level=Parallelizer.base_log_level)
            #     self.queues[self.id].init_flag.set()
            # if self.parent.on_main:
            #     for i, q in enumerate(self.queues):
            #         if not q.init_flag.is_set():
            #             self.parent.print("checking init flag on {i}".format(i=i), log_level=Parallelizer.base_log_level)
            #             wat = q.init_flag.wait(self.initialization_timeout)
            #             if not wat:
            #                 raise self.PoolError("Failed to initialize pool")

        @property
        def locations(self):
            """
            Returns the list of queue ids
            :return:
            :rtype:
            """
            return [x.id for x in self.queues]
        @property
        def location(self):
            """
            Returns the _current_ location
            :return:
            :rtype:
            """
            return self.id

        def send(self, data, loc, prepickled=False, **kwargs):
            """
            Sends the specified data to loc
            :param data:
            :type data:
            :param loc:
            :type loc:
            :param kwargs:
            :type kwargs:
            :return:
            :rtype:
            """

            queue = self.queues[loc].send_queue #type: mp.queues.Queue
            if loc == self.id:
                self.parent.print("Send: getting on {id}".format(id=self.id), log_level=Parallelizer.base_log_level)
                res = queue.get()
                self.parent.print("Send: got on {id}".format(id=self.id), log_level=Parallelizer.base_log_level)
                res = pickle.loads(res)
                return res
            else:
                self.parent.print("Send: putting {id} to {loc}".format(id=self.id, loc=loc), log_level=Parallelizer.base_log_level)
                if not prepickled:
                    data = pickle.dumps(data, protocol=pickle.HIGHEST_PROTOCOL)
                queue.put(data)
                self.parent.print("Send: put on {id} to {loc}".format(id=self.id, loc=loc), log_level=Parallelizer.base_log_level)
                return data
        def receive(self, data, loc, prepickled=False, **kwargs):
            """
            Receives the specified data from loc
            :param data:
            :type data:
            :param loc:
            :type loc:
            :param kwargs:
            :type kwargs:
            :return:
            :rtype:
            """
            queue = self.queues[loc].receive_queue
            if loc != self.id:
                self.parent.print("Recv: getting on {id} from {loc}".format(id=self.id, loc=loc), log_level=Parallelizer.base_log_level)
                res = queue.get()
                res = pickle.loads(res)
                self.parent.print("Recv: got on {id} from {loc}".format(id=self.id, loc=loc), log_level=Parallelizer.base_log_level)
                return res
            else:
                self.parent.print("Recv: putting on {id}".format(id=self.id), log_level=Parallelizer.base_log_level)
                if not prepickled:
                    data = pickle.dumps(data, protocol=pickle.HIGHEST_PROTOCOL)
                queue.put(data)
                self.parent.print("Recv: put on {id}".format(id=self.id), log_level=Parallelizer.base_log_level)
                return data

        def get_subcomm(self, idx):
            return type(self)(
                self.parent,
                self.id,
                tuple(self.queues[i] for i in idx),
                initialization_timeout=self.initialization_timeout,
                group=None if self.group is None else tuple(self.group[i] for i in idx)
            )

    _is_worker=False # global flag to be overridden
    def __init__(self,
                 worker=False,
                 pool:mp.Pool=None,
                 context=None,
                 manager=None,
                 logger=None,
                 contract=None,
                 comm=None,
                 rank=None,
                 allow_restart=True,
                 initialization_timeout=.5,
                 initialization_function=None,
                 initialization_args=None,
                 initialization_kwargs=None,
                 **kwargs
                 ):
        self.initialization_timeout=initialization_timeout
        super().__init__(logger=logger, contract=contract,
                         initialization_function=initialization_function,
                         initialization_args=initialization_args,
                         initialization_kwargs=initialization_kwargs,
                         )
        self.opts = kwargs
        self.pool:mp_pool.Pool = pool
        self.worker = worker
        self.ctx = context
        self.manager:mp.Manager = manager
        self._comm = comm
        self._id = rank
        self.nproc = None
        self.allow_restart = allow_restart

    def get_nprocs(self):
        return self.nproc
    def get_id(self):
        if self._id is None:
            return self.comm.id
        else:
            return self._id

    @property
    def comm(self):
        """
        Returns the communicator used by the paralellizer
        :return:
        :rtype: MultiprocessingParallelizer.PoolCommunicator
        """
        if self._comm is None:
            self.logger.log_print("initializing PoolCommunicators...", log_level=self.logger.LogLevel.MoreDebug)
            comm_list = [
                self.PoolCommunicator(self, i, self.queues, initialization_timeout=self.initialization_timeout) for
                i in range(0, self.nproc)
            ]
            self.logger.log_print("got comm group {g}", g=comm_list, log_level=self.logger.LogLevel.MoreDebug)
            self._comm = comm_list[0]
            self._comm.group = comm_list
        return self._comm
    @comm.setter
    def comm(self, c):
        self._comm = c

    # def to_state(self, serializer=None):
    #     return self.__getstate__()
    # @classmethod
    # def from_state(cls, state, serializer=None):
    #     return cls(
    #         worker=state['worker']
    #     )
    def __getstate__(self):
        # most things don't need to be mapped over...
        state = self.__dict__.copy()
        state['pool'] = None
        state['worker'] = True
        state['manager'] = None
        state['_default_stack'] = None
        # state['comm'] = None
        state['_comm'] = None
        state['queues'] = None
        state['_pid'] = None
        # state['_active_sentinel'] = 0
        # state['_id'] = self.id
        # state['_par_registry'] = None
        return state
    def __setstate__(self, state):
        self.__dict__.update(state)
        # allow for better syncing...
        if self.uid in self.parallelizer_registry:
            parent = self.lookup(self.uid)
            if parent is not None:
                self.__dict__.update(parent.__dict__)
            # print("?", parent)
        # else:
        #     print(":o", self, list(self.parallelizer_registry.values()))
        #     self.register(self.uid)

    # def broadcast(self, data, locs=None, prepickled=False, **kwargs):
    #     """
    #     Sends the same data to all processes
    #
    #     :param data:
    #     :type data:
    #     :param kwargs:
    #     :type kwargs:
    #     :return:
    #     :rtype:
    #     """
    #     if self.on_main:
    #         if not prepickled:
    #             pick = pickle.dumps(data, protocol=pickle.HIGHEST_PROTOCOL)
    #         else:
    #             pick = data
    #             data = pickle.loads(data)
    #         super().broadcast(pick, locs=locs, prepickled=True, **kwargs)
    #         return data
    #     else:
    #         return super().broadcast(data, locs=locs, prepickled=prepickled, **kwargs)

    @staticmethod
    def _run(runner, comm:PoolCommunicator, args, kwargs, main_kwargs=None):
        """
        Static runner function that just dispatches methods out to
        cores
        :param runner:
        :type runner:
        :param comm:
        :type comm:
        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        if runner is None: #quick ignore
            return None

        self = comm.parent
        self._comm = comm # makes a cyclic dependency...not sure how best to fix that
        self.worker = comm.id > 0
        with self:
            # print(self._active_sentinel, list(self.parallelizer_registry.values()))
            if self.on_main:
                self.print(
                    "Starting Parallelizer with {runner} over processor group {grp}".format(
                        runner=runner,
                        grp=self.comm.locations
                    ),
                    log_level=Parallelizer.base_log_level
                )
            else:
                self.print(
                    "starting process {p} at {t}".format(
                        p=self.pid,
                        t=time.ctime()
                    ),
                    log_level=Parallelizer.base_log_level
                )
            self._comm.initialize()
            kwargs['parallelizer'] = comm.parent
            if main_kwargs is None:
                main_kwargs = {}
            if self.on_main:
                return runner(*args, **main_kwargs, **kwargs)
            else:
                try:
                    return runner(*args, **main_kwargs, **kwargs)
                except Exception as e:
                    import traceback as tb
                    self.print(tb.format_exc())
                    comm.send(self.ReceivedError(e), 0)
                    raise

    def apply(self, func, *args, comm=None, main_kwargs=None, cleanup=True, **kwargs):
        """
        Applies func to args in parallel on all of the processes

        :param func:
        :type func:
        :param args:
        :type args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        if comm is None:
            comm = self.comm
        elif not isinstance(comm, SendRecieveParallelizer.SendReceieveCommunicator):
            comm = self.comm.get_subcomm(comm)

        _comm = self.comm
        try:
            self.comm = comm

            # since comm might be smaller than the pool
            # we need to send "ignore" flags to those pools
            if comm.group is None:
                raise ValueError("{} needs to have a processor group to parallelize over".format(
                    comm
                ))
            group_map = {g.id:g for g in comm.group}
            if 0 not in group_map:
                raise ValueError("{} needs to have the main process in it as designed".format(
                    comm
                ))

            mapping = [
                [
                    func,
                    group_map[i],
                    args,
                    kwargs
                ] if i in group_map else [
                    None,
                    None,
                    None,
                    None
                ] for i in range(1, self.nproc)
            ]
            pool = self.pool #type: mp_pool.Pool
            if cleanup:
                self.comm.reset()
            subsidiary = pool.starmap_async(self._run, mapping)
            # pool._worker_handler.join()
            # pool._task_handler.join()
            # pool._help_stuff_finish(
            #     pool._inqueue,
            #
            # )
            # while not (
            #          pool._taskqueue.empty()
            #          and pool._inqueue.empty()
            # ):
            #     time.sleep(.05)
            # if self.initialization_timeout is not None:
            #     time.sleep(self.initialization_timeout)
            try:
                main = self._run(func, comm, args, kwargs, main_kwargs=main_kwargs)
            except self.PoolCommunicator.PoolError:
                # check for errors on subsidiary...
                subsidiary.get(timeout=self.initialization_timeout)
                raise
            subs = subsidiary.get() # just to effect a wait
        finally:
            self.comm = _comm
        return main

    def _get_pool(self,
                  manager: mp.Manager,
                  **kwargs
                  ) -> mp.pool.Pool:
        if 'processes' not in kwargs:
            kwargs['processes'] = mp.cpu_count() - 1
        if 'initializer' not in kwargs:
            kwargs['initializer'] = functools.partial(
                self._initialize_worker,
                self.initialization_function
            )
        return manager.Pool(**kwargs)
    @staticmethod
    def get_pool_context(pool):
        return pool._ctx # I don't like doing this but seems like only way?
    @staticmethod
    def get_pool_nprocs(pool):
        return pool._processes  # see above

    def _reset_mp_caches(self):
        self.pool = None
        self.queues = None
        self.comm = None
    @classmethod
    def _initialize_worker(cls, initializer=None, *etc):
        """
        Sets a flag so that worker processes
        can know immediately that they are workers
        """
        if initializer is not None:
            initializer()
        cls._is_worker = True
    def set_initializer(self, func, *args, **kwargs):
        self.initialization_function = functools.partial(func, *args, **kwargs)
        self.initialization_function()
        self.pool.map(
            self._initialize_worker,
            [self.initialization_function] * (self.nproc)
        )
        self.pool.initializer = functools.partial(
            self._initialize_worker,
            self.initialization_function
        )

    def initialize(self, allow_restart=None):
        if not self.worker:
            if self.pool is None:
                self.print("Initializing pool...", log_level=Parallelizer.base_log_level)
                if self.ctx is None:
                    self.ctx = mp.get_context() # get the default context
                self.pool:mp.Pool = self._get_pool(self.ctx, **self.opts)
            elif self.ctx is None:
                self.ctx = self.get_pool_context(self.pool)
            if self.manager is None:
                self.manager = mp.Manager()
            allow_restart = self.allow_restart if allow_restart is None else allow_restart
            if allow_restart:
                try:
                    self.pool.__enter__()
                except ValueError:
                    # fix poos
                    self._reset_mp_caches()
                    return self.initialize(allow_restart=False)
            else:
                self.pool.__enter__()
            self.nproc = self.get_pool_nprocs(self.pool)
            self.pool.map(
                self._initialize_worker,
                [self.initialization_function] * (self.nproc+1)
            )
            if self.initialization_function is not None:
                self.initialization_function()
            # just to be safe
            self._is_worker = False
            self.worker = False
            self.queues = [
                self.SendRecvQueuePair(i, self.manager)
                for i in range(0, self.nproc)
            ]

    def finalize(self, exc_type, exc_val, exc_tb):
        if not self.worker:
            if self.pool is not None:
                self.pool.__exit__(exc_type, exc_val, exc_tb)
            self.queues = None
            self._comm = None
    @property
    def on_main(self):
        return not self.worker

    @classmethod
    def from_config(cls, **kw):
        return cls(**kw)
Parallelizer.mode_map['multiprocessing'] = MultiprocessingParallelizer

class MPIParallelizer(SendRecieveParallelizer):
    """
    Parallelizes using `mpi4py`
    """

    class MPICommunicator(SendRecieveParallelizer.SendReceieveCommunicator):
        """
        A base class that provides an interface for
        sending/receiving data from a specific subprocesses
        """

        def __init__(self, parent, mpi_comm, api):
            self.parent = parent
            self.comm = mpi_comm
            self.api = api
        @property
        def type_map(self):
            """
            Kinda hacky type map...hopefully
            sufficient to just have a few core numeric types?
            """
            return {
                'int32': self.api.INT,
                'int64': self.api.LONG,
                'float32': self.api.FLOAT,
                'float64': self.api.DOUBLE
            }
        def get_mpi_type(self, dtype):
            """
            Gets the MPI datatype for the numpy
            dtype
            """
            tm = self.type_map
            key = dtype.name
            if key not in tm:
                raise KeyError(
                    "mapping from {} to MPI datatype not included...".format(dtype)
                    )
            return tm[key]
        @property
        def nprocs(self):
            """
            Returns the number of available processes
            :return:
            :rtype:
            """
            return self.comm.Get_size()
        @property
        def locations(self):
            """
            Returns the list of locations known by the
            `SendReceieverCommunicator`
            :return:
            :rtype:
            """
            return range(self.comm.Get_size())
        @property
        def location(self):
            """
            Returns the _current_ location
            :return:
            :rtype:
            """
            return self.comm.Get_rank()
        def send(self, data, loc, **kwargs):
            """
            Sends the specified data to loc
            :param data:
            :type data:
            :param loc:
            :type loc:
            :param kwargs:
            :type kwargs:
            :return:
            :rtype:
            """
            if loc == self.location:
                return self.comm.recv(source=self.api.ANY_SOURCE)
            else:
                self.comm.send(data, dest=loc)
        def receive(self, data, loc, root=0, **kwargs):
            """
            Receives the specified data from loc
            :param data:
            :type data:
            :param loc:
            :type loc:
            :param kwargs:
            :type kwargs:
            :return:
            :rtype:
            """
            if loc != self.location:
                self.comm.send(self.location, dest=loc)
                return self.comm.recv(source=loc)
            else:
                where_to = self.comm.recv(source=self.api.ANY_SOURCE)
                self.comm.send(data, dest=where_to)
        def broadcast(self, data, root=0, **kwargs):
            """
            Sends the same data to all processes

            :param data:
            :type data:
            :param kwargs:
            :type kwargs:
            :return:
            :rtype:
            """
            if isinstance(data, np.ndarray):
                return self.comm.Bcast(data, root=root)
            else:
                return self.comm.bcast(data, root=root)
        def scatter_obj(self, data, root=0, **kwargs):
            """
            Scatters data to the different MPI ranks using a series
            of send calls.
            This is the default for anything except numpy arrays.

            :param data:
            :type data:
            :param root:
            :type root:
            :param kwargs:
            :type kwargs:
            :return:
            :rtype:
            """
            if self.location == root:
                locs = list(self.locations)  # gotta be safe
                nlocs = len(locs)
                chunk_size = len(data) // nlocs
                chunk_sizes = [chunk_size] * nlocs
                for i in range(len(data) - (chunk_size*nlocs)):
                    chunk_sizes[i] += 1

                s = chunk_sizes[0]
                main_data = data[:s]
                for i, b in zip(locs[1:], chunk_sizes[1:]):
                    chunk = data[s:s + b]
                    self.send(chunk, i, **kwargs)
                    s += b
                return main_data
        def scatter(self, data, root=0, shape=None, dtype=None, **kwargs):
            """
            Performs a scatter of data to the different
            available parallelizer processes.
            *NOTE:* unlike in the MPI case, `data` does not
            need to be evenly divisible by the number of available
            processes

            :param data:
            :type data:
            :param kwargs:
            :type kwargs:
            :return:
            :rtype:
            """

            if isinstance(data, np.ndarray) or data is None:
                if root == self.location and data is None:
                    raise TypeError("'None' is not scatterable. Try `broadcast`.")
                if root == self.location:
                    send_buf = np.ascontiguousarray(data)
                else:
                    send_buf = None
                if shape is None:
                    if root == self.location:
                        shape = data.shape
                    shape = self.broadcast(shape)
                if dtype is None:
                    if root == self.location:
                        dtype = data.dtype
                    dtype = self.broadcast(dtype)
                np.empty(shape, dtype=dtype)
                ranks = self.comm.Get_size()
                ndat = shape[0]
                block_size = ndat // ranks
                block_remainder = ndat - (block_size*ranks)
                if block_remainder == 0:
                    shape = (block_size,) + shape[1:]
                    recv_buf = np.empty(shape, dtype=dtype)
                    self.comm.Scatter(send_buf, recv_buf, root=root)
                    return recv_buf
                else:
                    # it turns out MPI4Py has only an unsophisticated
                    # implementation of Scatterv that explicitly needs the
                    # appropriate block offsets, since it basically expects
                    # to have a flattened form of the array (just like MPI)
                    block_sizes = [block_size]*ranks
                    for i in range(block_remainder):
                        block_sizes[i] += 1

                    recv_buf = np.empty((block_sizes[self.location],) + shape[1:], dtype=dtype)

                    block_offset = int(np.prod(shape[1:]))
                    block_sizes = np.array(block_sizes)*block_offset
                    block_offsets = np.concatenate([[0], np.cumsum(block_sizes)])[:-1]
                    # print(block_offsets, block_offset, block_sizes)

                    # self.parent.print("block sizes: {} offsets: {}".format(block_sizes, block_offsets))
                    self.comm.Scatterv(
                        [
                            send_buf,
                            block_sizes,
                            block_offsets,
                            self.get_mpi_type(dtype)
                        ],
                        recv_buf,
                        root=root
                    )
                    self.parent.print("sending", send_buf, recv_buf, self.get_mpi_type(dtype).name,
                                      log_level=Parallelizer.base_log_level)
                    return recv_buf
            else:
                return self.scatter_obj(data, root=root, **kwargs)
        def gather_obj(self, data, root=0, **kwargs):
            if self.location == root:
                locs = list(self.locations)  # gotta be safe
                nlocs = len(locs)  # we reserve space for the main thread
                recv = [None] * nlocs
                recv[0] = data
                for n, i in enumerate(locs[1:]):
                    recv[n + 1] = self.receive(data, i, **kwargs)
                return recv
            else:
                return self.receive(data, self.location, **kwargs)  # effectively a send...
        def gather(self, data, root=0, shape=None, dtype=None, **kwargs):
            """
            Performs a gather from the different
            available parallelizer processes.

            :param data:
            :type data:
            :param kwargs:
            :type kwargs:
            :return:
            :rtype:
            """

            if isinstance(data, np.ndarray):
                # make sure everyone knows what's up
                send_buf = np.ascontiguousarray(data)
                if shape is None:
                    block_sizes = self.gather(data.shape[0], root=root)
                    if block_sizes is not None:
                        ndat = int(np.sum(block_sizes))
                        if root == self.location:
                            shape = (ndat,) + data.shape[1:]
                    shape = self.broadcast(shape)
                if dtype is None:
                    if root == self.location:
                        dtype = data.dtype
                    dtype = self.broadcast(dtype)
                # otherwise send shit around
                ranks = self.comm.Get_size()
                ndat = shape[0]
                block_size = ndat // ranks
                block_remainder = ndat - (block_size*ranks)
                if block_remainder == 0:
                    if root == self.location:
                        recv_buf = np.empty(shape, dtype=dtype)
                    else:
                        recv_buf = None
                    self.comm.Gather(send_buf, recv_buf, root=root)
                    return recv_buf
                else:
                    block_sizes = [block_size] * ranks
                    for i in range(block_remainder):
                        block_sizes[i] += 1
                    block_offset = int(np.prod(shape[1:]))
                    block_sizes = np.array(block_sizes)*block_offset
                    block_offsets = np.concatenate([[0], np.cumsum(block_sizes)])[:-1]
                    if root == self.location:
                        recv_buf = np.empty(shape, dtype=dtype)
                    else:
                        recv_buf = None
                    self.comm.Gatherv(
                        send_buf,
                        [
                            recv_buf,
                            block_sizes,
                            block_offsets,
                            self.get_mpi_type(dtype)
                        ],
                        root=root
                    )

                    return recv_buf
            else:
                return self.gather_obj(data, root=root, **kwargs)

    def __init__(self, root=0, comm=None, contract=None, logger=None):
        super().__init__(contract=contract, logger=logger)

        from mpi4py import MPI as api
        self.api = api

        if comm is None:
            comm = api.COMM_WORLD
        self.root = root
        self._comm = self.MPICommunicator(self, comm, api)
        self.world_size = None
        self.world_rank = None

    def get_nprocs(self):
        return self._comm.nprocs
    def get_id(self):
        return self._comm.location

    def initialize(self):
        # handled by mpi4py?
        pass
    def finalize(self, exc_type, exc_val, exc_tb):
        # handled by mpi4py?
        pass

    @property
    def comm(self):
        """
        Returns the communicator used by the paralellizer
        :return:
        :rtype: MPIParallelizer.MPICommunicator
        """
        return self._comm

    @property
    def on_main(self):
        return self.comm.location == 0

    def broadcast(self, data, **kwargs):
        """
        Sends the same data to all processes

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        if self.contract is not None:
            self.contract.handle_call(self, "broadcast")
        return self.comm.broadcast(data, root=self.root, **kwargs)
    def scatter(self, data, shape=None, **kwargs):
        """
        Performs a scatter of data to the different
        available parallelizer processes.
        *NOTE:* unlike in the MPI case, `data` does not
        need to be evenly divisible by the number of available
        processes

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        if self.contract is not None:
            self.contract.handle_call(self, "scatter")
        return self.comm.scatter(data, root=self.root, shape=shape, **kwargs)
    def gather(self, data, shape=None, **kwargs):
        """
        Performs a gather of data from the different
        available parallelizer processes

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        if self.contract is not None:
            self.contract.handle_call(self, "gather")
        return self.comm.gather(data, root=self.root, shape=shape, **kwargs)
    def map(self, func, data, input_shape=None, output_shape=None, **kwargs):
        """
        Performs a parallel map of function over
        the held data on different processes

        :param function:
        :type function:
        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        sub_data = self.scatter(data, shape=input_shape, **kwargs)
        res = func(sub_data)
        return self.gather(res, shape=output_shape, **kwargs)

    def apply(self, func, *args, cleanup=True, **kwargs):
        """
        Applies func to args in parallel on all of the processes.
        For MPI, since jobs are always started with mpirun, this
        is just a regular apply

        :param func:
        :type func:
        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return func(*args, parallelizer=self, **kwargs)

    @classmethod
    def from_config(cls, **kw):
        return cls(**kw)
Parallelizer.mode_map['mpi'] = MPIParallelizer

class SerialNonParallelizer(Parallelizer):
    """
    Totally serial evaluation for cases where no parallelism
    is provide
    """

    def get_nprocs(self):
        return 1
    def get_id(self):
        return 0

    def initialize(self):
        """
        Initializes a parallelizer
        if necessary
        :return:
        :rtype:
        """
        if self.initialization_function is not None:
            self.initialization_function()
        pass
    def finalize(self, exc_type, exc_val, exc_tb):
        """
        Finalizes a parallelizer (if necessary)
        if necessary
        :return:
        :rtype:
        """
        pass
    @property
    def on_main(self):
        """
        Returns whether or not the executing process is the main
        process or not
        :return:
        :rtype:
        """
        return True
    def send(self, data, loc, **kwargs):
        """
        A no-op

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return data
    def receive(self, data, loc, **kwargs):
        """
        A no-op

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return data
    def broadcast(self, data, **kwargs):
        """
        A no-op

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return data
    def scatter(self, data, **kwargs):
        """
        A no-op

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return data
    def gather(self, data, **kwargs):
        """
        A no-op

        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return [data]

    def map(self, function, data, extra_args=None, extra_kwargs=None,
            vectorized=True,
            aggregate=False,
            **kwargs):
        """
        Performs a serial map of the function over
        the passed data

        :param function:
        :type function:
        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        if extra_args is not None or extra_kwargs is not None:
            if extra_args is None:
                extra_args=()
            if extra_kwargs is None:
                extra_kwargs={}
            function = lambda a, fn=function, ea=extra_args, ek=extra_kwargs: fn(a, *ea, **ek)

        if vectorized:
            res = function(data, **kwargs)
        else:
            res = list(map(function, data, **kwargs))

        return res

    def starmap(self, function, data, extra_args=None, extra_kwargs=None, **kwargs):
        """
        Performs a serial map with unpacking of the function over
        the passed data

        :param function:
        :type function:
        :param data:
        :type data:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        if extra_args is None:
            extra_args=()
        if extra_kwargs is None:
            extra_kwargs={}
        function = lambda a, fn=function, ea=extra_args, ek=extra_kwargs: fn(*a, *ea, **ek)

        return list(map(function, data, **kwargs))

    def apply(self, func, *args, comm=None, main_kwargs=None, cleanup=True, **kwargs):
        kwargs['parallelizer'] = self
        if main_kwargs is None:
            main_kwargs = {}
        return func(*args, **main_kwargs, **kwargs)

    def wait(self):
        """
        No need to wait when you're in a serial environment
        :return:
        :rtype:
        """

        pass






"""
The coordinate mats class defines an architecture to mats coordinates
"""
import numpy as np
from collections import namedtuple

######################################################################################################
##
##                                   GeometricTransformation Class
##
######################################################################################################

from .TransformationFunction import TransformationFunction
from .AffineTransform import AffineTransform
from .RotationTransform import RotationTransform
from .ScalingTransform import ScalingTransform
from .TranslationTransform import TranslationTransform

__all__ = [
    "GeometricTransformation"
]

__reload_hook__ = ['.TransformationFunction', ".AffineTransform", ".RotationTransform", ".ScalingTransform", ".TranslationTransform"]

class GeometricTransformation:
    """
    The GeometricTransformation class provides a simple, general way to represent a
    compound coordinate transformation.
    In general, it's basically just a wrapper chaining together a number of TransformationFunctions.
    """

    def __init__(self, *transforms):
        if len(transforms) == 0:
            transforms = (ScalingTransform([1, 1, 1]),)
        self._transform_list = tuple(self.parse_transform(tf) for tf in transforms)
        self._trans = None
        self.condense_transforms()

    @property
    def is_affine(self):
        return isinstance(self._trans, AffineTransform)
    @property
    def transformation_function(self):
        """

        :return:
        :rtype: TransformationFunction
        """
        return self._trans
    @property
    def transforms(self):
        return self._transform_list

    def apply(self, coords, shift=True):
        tfunc = self.transformation_function #type: TransformationFunction
        return tfunc.operate(coords, shift=shift)
    def __call__(self, coords, shift=True):
        if isinstance(coords, (TransformationFunction, GeometricTransformation)):
            return type(self)(self, coords) # just constructing a new transform...
        else:
            return self.apply(coords)

    def condense_transforms(self):
        self._trans = self._transform_list[-1]
        for t in self._transform_list[::-1][1:]:
            self._trans = t.merge(self._trans)

    @property
    def inverse(self):
        return type(self)(self._trans.inverse)

    @staticmethod
    def parse_transform(tf):
        """
        Provides a way to "tag" a transformation
        :param tf:
        :type tf:
        :return:
        :rtype:
        """

        if isinstance(tf, TransformationFunction) or hasattr(tf, 'operate'):
            return tf
        elif isinstance(tf, GeometricTransformation) or hasattr(tf, 'transformation_function'):
            return tf.transformation_function
        elif isinstance(tf, str):
            # we can define some convenient transformation syntax, maybe
            raise NotImplementedError("String syntax for transformations still isn't implemented. Sorry about that.")
        else:
            tf = np.asarray(tf)
            if tf.ndim == 1:
                tf = TranslationTransform(tf)
            elif tf.shape[0] == 3 and tf.shape[1] == 1:
                tf = ScalingTransform(tf.flatten())
            else:
                tf = AffineTransform(tf)
            return tf




import uuid

from .JHTML import HTML

__all__ = [
    "JSMol"
]

__reload_hooks__ = [".JHTML"]

class JSMol:
    class Applet(HTML.Div):
        version = "16.3.7.6"
        jsmol_source = f"https://cdn.jsdelivr.net/gh/b3m2a1/jsmol-cdn@{version}/jsmol/JSmol.min.js"
        jmol2_source = f"https://cdn.jsdelivr.net/gh/b3m2a1/jsmol-cdn@{version}/jsmol/js/Jmol2.js"
        patch_script = f"""
        if (typeof Jmol._patched === 'undefined') {{
            Jmol._patched = true;
            jmolInitialize('https://cdn.jsdelivr.net/gh/b3m2a1/jsmol-cdn@{version}/jsmol/');
        }};
        """
        @classmethod
        def load_applet_script(cls, id, loader):
            return HTML.Script(src=cls.jsmol_source,
                               onload=f"""
(function() {{
   $.getScript('{cls.jmol2_source}').then(
   () => {{
       {cls.patch_script}
       let loaded = false;
        if (!loaded) {{
            loaded = true;
            let applet = {loader};
             document.getElementById('{id}').innerHTML = applet._code;
        }}
    }})
}})();
"""
                               )

        def __init__(self, *model_etc, width='500px', height='500px',
                     animate=False, vibrate=False,
                     load_script=None,
                     suffix=None,
                     dynamic_loading=None,
                     **attrs):
            if suffix is None:
                suffix = str(uuid.uuid4())[:6].replace("-", "")
            self.suffix = suffix
            self.id = "jsmol-applet-" + self.suffix
            if len(model_etc) > 0 and isinstance(model_etc[0], str):
                model_file = model_etc[0]
                rest = model_etc[1:]
            else:
                model_file = None
                rest = model_etc

            if load_script is None:
                load_script = []
            if isinstance(load_script, str):
                load_script = [load_script]
            load_script = list(load_script)

            if animate:
                load_script.extend(["anim mode palindrome", "anim on"])
            elif vibrate:
                load_script.append("vibration on")

            if dynamic_loading is None:
                from ..Jupyter.JHTML import JupyterAPIs
                dynamic_loading = JupyterAPIs().in_jupyter_environment()

            self.dynamic_loading = dynamic_loading

            self.load_script = load_script
            elems = [self.create_applet(model_file)] + list(rest)
            super().__init__(*elems, id=self.id, width=width, height=height, **attrs)

        @property
        def applet_target(self):
            return f"_{self.suffix}"
        def prep_load_script(self):
            return '; '.join(self.load_script)
        def create_applet(self, model_file):
            targ = self.applet_target
            load_script = self.prep_load_script()
            if model_file is None:
                loader = f"jmolApplet(400, 'load {model_file}; {load_script}', '{targ}')"
            elif (
                    model_file.startswith("https://")
                    or model_file.startswith("file://")
                    or model_file.startswith("http://")
            ):
                loader = f"jmolApplet(400, 'load {model_file}; {load_script}', '{targ}')"
            else:
                loader = f"jmolAppletInline(400, `{model_file}`, '{load_script}', '{targ}')"

            kill_id = "tmp-" + str(uuid.uuid4())[:10]
            load_script = self.load_applet_script(self.id, loader)

            if self.dynamic_loading:
                load_script = load_script.tostring().replace("`", "\`")
                return HTML.Image(
                        src='data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7',
                        id=kill_id,
                        onload=f"""
                            (function() {{
                                document.getElementById('{kill_id}').remove();
                                const frag = document.createRange().createContextualFragment(`{load_script}`);
                                document.head.appendChild(frag);
                            }})()"""
                    )
            else:
                return load_script

        def show(self):
            return self.display()


import json

__all__ = [
    "SceneJSON"
]

class SceneJSON:
    """
    JSON interchange format to use with Mathematica
    """
    class Primitive:
        def __init__(self, tag, *children, **attrs):
            self.tag = tag
            if len(children) == 1 and isinstance(children[0], (list, tuple)):
                children = children[0]
            self.children = list(children)
            self.attrs = attrs
        def to_json(self):
            return dict(
                tag=self.tag,
                children=[
                    e.to_json()
                    for e in self.children
                ],
                attrs=self.attrs
            )
        def tostring(self, **opts):
            return json.dumps(self.to_json(), **opts)
        def dump(self, file, **opts):
            if hasattr(file, 'write'):
                return json.dump(self.to_json(), file, **opts)
            else:
                with open(file, 'w+') as file:
                    return json.dump(self.to_json(), file, **opts)

    class TagElement(Primitive):
        tag = None
        def __init__(self, *children, **attrs):
            super().__init__(self.tag, *children, **attrs)

    class Graphics3D(TagElement): tag = "graphics3d"
    class Graphics(TagElement): tag = "graphics"
    class Animation(TagElement): tag = 'animation'
    class Scene(TagElement): tag = 'scene'
    class Circle(TagElement): tag = 'circle'
    class Line(TagElement): tag = 'line'
    class Rectangle(TagElement): tag = 'rectangle'
    class Polygon(TagElement): tag = 'polygon'
    class Disk(TagElement): tag = 'disk'
    class Cone(TagElement): tag = 'cone'
    class Sphere(TagElement): tag = 'sphere'
    class Cuboid(TagElement): tag = 'cuboid'
    class Cylinder(TagElement): tag = 'cylinder'
    class Text(TagElement): tag = 'text'


"""
Provides functionality for managing large sets of options
"""

import os, inspect
from typing import Callable
from . import core

__all__ = [
    "OptionsSet",
    "OptionsMethodDispatch"
]

class OptionsSet:
    """
    Provides a helpful manager for those cases where
    there are way too many options and we need to filter
    them across subclasses and things
    """

    def __init__(self, *d, **ops):
        if len(d) > 0:
            if isinstance(d[0], dict):
                self.ops = d[0]
                self.ops.update(ops)
            else:
                self.ops = dict(d, **ops)
        else:
            self.ops = ops
    def __getitem__(self, item):
        return self.ops[item]
    def __setitem__(self, key, value):
        self.ops[key] = value
    def __delitem__(self, item):
        del self.ops[item]
    def __getattr__(self, item):
        return self.ops[item]
    def __setattr__(self, key, value):
        if key == "ops":
            super().__setattr__(key, value)
        else:
            self.ops[key] = value
    def __delattr__(self, item):
        del self.ops[item]
    def __hasattr__(self, key):
        return key in self.ops
    def update(self, **ops):
        self.ops.update(**ops)

    def keys(self):
        return self.ops.keys()
    def items(self):
        return self.ops.items()

    def save(self, file, mode=None, attribute=None):
        self.serialize(file)
    @classmethod
    def load(cls, file, mode=None, attribute=None):
        cls(cls.deserialize(file, mode=mode, attribute=attribute))

    def extract_kwarg_keys(self, obj):
        args, _, _, defaults, _, _, _  = inspect.getfullargspec(obj)
        if args is None:
            return None
        ndef = len(defaults) if defaults is not None else 0
        return tuple(args[-ndef:])
    def get_props(self, obj):
        if isinstance(obj, (list, tuple)):
            return sum(
                (self.get_props(o) for o in obj),
                ()
            )

        try:
            props = obj.__props__
        except AttributeError:
            try:
                annotations = obj.__annotations__
            except AttributeError:
                annotations = {}
            if len(annotations) == 0:
                props = self.extract_kwarg_keys(obj)
            else:
                props = tuple(annotations.keys())

        if props is None:
            raise AttributeError("{}: object {} needs props to filter against".format(
                type(self).__name__,
                self
            ))
        return props

    def bind(self, obj, props=None):
        for k,v in self.filter(obj, props=props).items():
            setattr(obj, k, self.ops[k])
    def filter(self, obj, props=None):
        if props is None:
            props = self.get_props(obj)
        ops = self.ops
        return {k:ops[k] for k in ops.keys() & set(props)}
    def exclude(self, obj, props=None):
        if props is None:
            props = self.get_props(obj)
        ops = self.ops
        return {k:ops[k] for k in ops.keys() - set(props)}
    def split(self, obj, props=None):
        if props is None:
            props = self.get_props(obj)
        return self.filter(obj, props=props), self.exclude(obj, props=props)


class OptionsMethodDispatch:
    def __init__(self,
                 methods_table:'dict|Callable[[], dict]',
                 attributes_map=None,
                 default_method=None,
                 methods_enum=None,
                 case_insensitive=True,
                 allow_custom_methods=True,
                 ignore_bad_enum_keys=False,
                 method_key='method'):
        if not hasattr(methods_table, 'items'):
            self.methods_table_generator = methods_table
            self.methods_table = {}
        else:
            self.methods_table = methods_table
            self.methods_table_generator = None
        self.attributes_map = attributes_map
        self.case_insensitive = case_insensitive
        self.method_key = method_key
        self.methods_enum = methods_enum
        self.default_method = default_method
        self.allow_custom_methods = allow_custom_methods
        self.ignore_bad_enum_keys = ignore_bad_enum_keys

    def register(self, method_name, method, base_attributes=None):
        self.methods_table[method_name] = method
        if base_attributes is not None:
            if self.attributes_map is None:
                self.attributes_map = {}
            if isinstance(base_attributes, str):
                base_attributes = (base_attributes,)
            self.attributes_map[tuple(base_attributes)] = method_name

    def load_methods_table(self):
        if self.methods_table_generator is not None:
            return dict(self.methods_table_generator(), **self.methods_table)
        else:
            return self.methods_table

    def _lookup_method(self, method, opts):
        if method is None and self.attributes_map is not None:
            for params, method_name in sorted(
                    self.attributes_map.items(),
                    key=lambda kt: -len(kt[0])
            ):
                if params is not None and all(p in opts for p in params):
                    return method_name

        return method

    def prep_method_spec(self, method_spec):
        if isinstance(method_spec, str) or core.is_default(method_spec):
            opts = {}
            method = method_spec
        elif hasattr(method_spec, 'name') and hasattr(method_spec, 'value'): # enum
            method = method_spec.value
            opts = {}
        elif core.is_dict_like(method_spec):
            opts = method_spec.copy()
            method = self._lookup_method(opts.pop(self.method_key, None), opts)
        elif callable(method_spec):
            method = method_spec
            opts = {}
        else:
            method, opts = method_spec
            method = self._lookup_method(method, opts)

        return method, opts

    def resolve(self, method_spec):
        method, opts = self.prep_method_spec(method_spec)
        if self.allow_custom_methods and callable(method):
            return method, opts

        methods_table = self.load_methods_table()
        if (
                self.methods_enum is not None
                and not core.is_default(method, allow_None=True)
                and method not in methods_table
        ):
            if self.case_insensitive and isinstance(method, str):
                try:
                    method = self.methods_enum(method)
                except ValueError:
                    if self.ignore_bad_enum_keys:
                        try:
                            method = self.methods_enum(method)
                        except ValueError:
                            ...
                    else:
                        method = self.methods_enum(method.lower())
            else:
                if self.ignore_bad_enum_keys:
                    try:
                        method = self.methods_enum(method)
                    except ValueError:
                        ...
                else:
                    method = self.methods_enum(method)

            if hasattr(method, 'value'):
                return methods_table.get(
                    method,
                    methods_table.get(method.value, methods_table.get(self.default_method))
                ), opts
            else:
                return methods_table.get(
                method,
                methods_table.get(self.default_method)
            ), opts
        else:
            if self.case_insensitive and isinstance(method, str) and not method in methods_table:
                method = method.lower()

            return methods_table.get(
                method,
                methods_table.get(self.default_method)
            ), opts


import shutil, os, sys, subprocess, importlib, platform
try:
    from distutils.core import setup, Extension
except ModuleNotFoundError:
    class Extension:
        def __init__(self, *args, **kwargs):
            raise ImportError("the FFI framework uses `distutils`, which has since been removed")

__all__ = [
    "CLoader"
]

class CLoader:
    """
    A general loader for C++ extensions to python, based off of the kind of thing that I have had to do multiple times
    """

    def __init__(self,
                 lib_name,
                 lib_dir=None,
                 load_path=None,
                 src_ext='src',
                 libs_ext="libs",
                 description="An extension module",
                 version="1.0.0",
                 include_dirs=None,
                 runtime_dirs=None,
                 linked_libs=None,
                 macros=None,
                 extra_link_args=None,
                 extra_compile_args=None,
                 extra_objects=None,
                 source_files=None,
                 build_script=None,
                 requires_make=True,
                 out_dir=None,
                 cleanup_build=True,
                 recompile=False
                 ):
        if lib_dir is None:
            if os.path.isdir(lib_name):
                lib_dir = lib_name
                lib_name = os.path.basename(lib_name)
            else:
                raise ValueError("'lib_dir' cannot be None")
        self.lib_name = lib_name
        self.lib_dir = lib_dir
        self.lib_description = description
        self.lib_version = version

        self.load_path = [self.lib_dir] if load_path is None else load_path
        self.include_dirs = () if include_dirs is None else tuple(include_dirs)
        self.runtime_dirs = runtime_dirs
        self.linked_libs = () if linked_libs is None else tuple(linked_libs)
        self.extra_link_args = () if extra_link_args is None else tuple(extra_link_args)
        self.extra_compile_args = () if extra_compile_args is None else tuple(extra_compile_args)
        self.extra_objects = () if extra_objects is None else tuple(extra_objects)
        self.src_ext = src_ext
        self.libs_ext = libs_ext
        self.macros = () if macros is None else tuple(macros)
        self.source_files = (lib_name+'.cpp',) if source_files is None else source_files
        self.build_script = build_script
        self.requires_make = requires_make

        self.out_dir = out_dir
        self.cleanup_build = cleanup_build

        self.recompile = recompile
        self._lib = None

    def load(self):
        if self._lib is None:
            ext = None if self.recompile else self.find_extension()
            if ext is None:
                ext = self.compile_extension()
            if ext is None:
                raise ImportError("Couldn't load/compile extension {} at {}".format(
                    self.lib_name,
                    self.lib_dir
                ))
            try:
                sys.path.insert(0, os.path.dirname(ext))
                module = os.path.splitext(os.path.basename(ext))[0]
                self._lib = importlib.import_module(module, self.lib_name)#+"Lib")
            finally:
                sys.path.pop(0)

        return self._lib

    def find_extension(self):
        """
        Tries to find the extension in the top-level directory

        :return:
        :rtype:
        """

        for l in self.load_path:
            r = self.locate_lib(l)[0]
            if r is not None:
                return r

    def compile_extension(self):
        """
        Compiles and loads a C++ extension

        :return:
        :rtype:
        """

        self.make_required_libs()
        self.build_lib()
        ext = self.cleanup()
        return ext

    @property
    def src_dir(self):
        return os.path.join(self.lib_dir, self.src_ext)
    @property
    def lib_lib_dir(self):
        return os.path.join(self.lib_dir, self.libs_ext)

    def get_extension(self):
        """
        Gets the Extension module to be compiled

        :return:
        :rtype:
        """
        lib_lib_dir = os.path.abspath(self.lib_lib_dir)
        is_mac = platform.system() == 'Darwin'

        lib_dirs = [os.path.abspath(d) for d in self.include_dirs + (lib_lib_dir,)]
        runtime_dirs = lib_dirs if self.runtime_dirs is None else lib_dirs + self.runtime_dirs
        libbies = list(self.linked_libs) + [f[3:].split(".")[0] for f in self.make_required_libs()]
        # print("????", runtime_dirs)
        mroos = self.macros
        sources = self.source_files

        extra_link_args = list(self.extra_link_args)
        extra_compile_args = list(self.extra_compile_args)
        if is_mac:
            # extra_link_args.append('-Xlinker -rpath -Xlinker' + ":".join(lib_dirs))
            extra_link_args.append('-headerpad_max_install_names')
            extra_link_args.extend('-Wl,-rpath,'+l for l in lib_dirs)#.join(lib_dirs))

        module = Extension(
            self.lib_name,
            sources=list(sources),
            library_dirs=list(lib_dirs),
            runtime_library_dirs=list(runtime_dirs),
            include_dirs=[os.path.abspath(d) for d in self.include_dirs],
            libraries=libbies,
            define_macros=list(mroos),
            extra_objects=list(self.extra_objects),
            extra_link_args=extra_link_args,
            extra_compile_args=extra_compile_args,
            language="c++"
        )

        return module

    def configure_make_command(self, make_file):

        python_dir = make_file["python_dir"]
        compiler = make_file["compiler"]
        sources = make_file["sources"] if 'sources' in make_file else self.source_files
        compiler_flags = make_file["compiler_flags"] if 'compiler_flags' in make_file else self.extra_compile_args
        include_dirs = make_file["include_directories"] if 'include_directories' in make_file else self.include_dirs
        runtime_dirs = make_file["runtime_directories"] if 'runtime_directories' in make_file else self.runtime_dirs
        if runtime_dirs is None:
            runtime_dirs = include_dirs

        linker = make_file["linker"]
        link_flags = make_file["link_flags"] if 'link_flags' in make_file else self.extra_link_args
        linked_libs = make_file["linked_libs"] if 'linked_libs' in make_file else self.linked_libs
        macros = make_file["macros"] if 'macros' in make_file else self.macros

        shared_lib = make_file["library_name"] if 'library_name' in make_file else self.lib_name

        if 'build_dir' not in make_file:
            source_dir = os.path.dirname(sources[0])
            build_dir = os.path.join(source_dir, "build")
        else:
            build_dir = make_file['build_dir']
        os.makedirs(build_dir)

        objects = [os.path.join(build_dir, os.path.splitext(s)[0] + ".o") for s in sources]

        compiler_flags = ["-f" + flag for flag in compiler_flags]
        include_dirs = ["-I" + i for i in include_dirs]

        if compiler == "CC":
            rpath = "-Wl,-R,"
        else:
            rpath = "-Wl,-rpath,"
        link_flags = ["-Wl," + f for f in link_flags]
        if len(include_dirs) > 0:
            runtime_dirs = [rpath + ":".join(include_dirs)]
        else:
            runtime_dirs = []
        linked_libs = ["-l" + l for l in linked_libs]
        link_dirs = ["-L" + i for i in include_dirs]
        "$link_flags $ryn_lib_src/build/RynLib.o $link_libs -o $ryn_lib_src/RynLib.so"

        shared_lib = shared_lib + (".pydll" if sys.platform() == "win32" else ".so")

        make_cmd = [
                       [compiler] + compiler_flags + macros + include_dirs + ["-c", src] + ["-o", obj] for src, obj in
                       zip(sources, objects)
                   ] + [
                       [linker] + link_flags + runtime_dirs + objects + link_dirs + linked_libs + ["-o", shared_lib]
                   ]

        return make_cmd

    def custom_make(self, make_file, make_dir):
        """
        A way to call a custom make file either for building the helper lib or for building the proper lib

        :param make_file:
        :type make_file:
        :param make_dir:
        :type make_dir:
        :return:
        :rtype:
        """

        curdir = os.getcwd()

        if isinstance(make_file, str) and os.path.isfile(make_file):
            make_dir = os.path.dirname(make_file)
            make_file = os.path.basename(make_file)
            if os.path.splitext(make_file)[1] == ".sh":
                make_cmd = ["bash", make_file]
            else:
                make_cmd = ["make", "-f", make_file]
        elif isinstance(make_file, dict):
            # little way to call the make process without a build.sh file
            make_cmd = self.configure_make_command(make_file)
        else:
            if os.path.exists(os.path.join(make_dir, "Makefile")):
                make_cmd = ["make"]
            elif os.path.exists(os.path.join(make_dir, "build.sh")):
                make_cmd = ["bash", "build.sh"]
            else:
                raise IOError(
                    "Can't figure out which file in {} should be the makefile. Expected either Makefile or build.sh".format(
                        make_dir
                    )
                )

        try:
            os.chdir(make_dir)
            if isinstance(make_cmd[0], str):
                out = subprocess.check_output(make_cmd)
                if len(out) > 0:
                    print(out.decode())
            else:
                for cmd in make_cmd:
                    out = subprocess.check_output(cmd)
                    if len(out) > 0:
                        print(out.decode())
        except subprocess.CalledProcessError as e:
            print(e.output.decode())
            raise
        finally:
            os.chdir(curdir)

    def make_required_libs(self, library_types=(".so", ".pyd", ".dll")):
        """
        Makes any libs required by the current one

        :return:
        :rtype:
        """

        lib_files = []
        if self.requires_make:
            lib_d = os.path.abspath(self.lib_lib_dir)
            lib_pairs = {}

            if os.path.isdir(lib_d):
                for lib_f in os.listdir(lib_d):
                    lib = os.path.join(lib_d, lib_f)
                    if lib_f.startswith('lib') and any(lib_f.endswith(ext) for ext in library_types):
                        lib_dir = os.path.join(lib_d, lib_f[3:].split(".")[0])
                        lib_pairs[lib_dir] = lib_f
                    elif os.path.isdir(lib):
                        if lib not in lib_pairs:
                            lib_pairs[lib] = None

            for lib, f in lib_pairs.items():
                if f is None:
                    self.custom_make(self.requires_make, lib)

            # need to reload
            if os.path.isdir(lib_d):
                for lib_f in os.listdir(lib_d):
                    if lib_f.startswith('lib') and any(lib_f.endswith(ext) for ext in library_types):
                        lib_files.append(lib_f)

        return lib_files

    def build_lib(self):

        curdir = os.getcwd()

        src_dir = os.path.abspath(self.src_dir)
        module = self.get_extension()

        sysargv1 = sys.argv
        custom_build = self.build_script
        if custom_build:
            self.custom_make(custom_build, src_dir)
        else:
            try:
                sys.argv = ['build', 'build_ext', '--inplace']
                os.chdir(src_dir)

                setup(
                    name=self.lib_description,
                    version=self.lib_version,
                    description=self.lib_description,
                    ext_modules=[module]
                )
            finally:
                sys.argv = sysargv1
                os.chdir(curdir)

        if platform.system() == "Darwin": # mac fuckery
            built, _, _ = self.locate_lib()
            libbies = list(self.linked_libs) + [f[3:].split(".")[0] for f in self.make_required_libs()]
            for l in libbies:
                subprocess.check_output(['install_name_tool', '-change', "lib{}.so".format(l), "@rpath/lib{}.so".format(l), built])

    @classmethod
    def locate_library(cls, libname, roots, extensions, library_types=(".so", ".pyd", ".dll")):
        """
        Tries to locate the library file (if it exists)

        :return:
        :rtype:
        """

        target = libname
        built = None
        ext = ""

        # test all roots and all extensions
        for top in roots:
            for extension in extensions:
                root = os.path.join(top, extension)
                if os.path.exists(root):
                    for f in os.listdir(root):
                        for library_type in library_types:
                            if f.startswith(libname) and f.endswith(library_type):
                                ext = library_type
                                built = os.path.join(root, f)
                                target += ext
                                break
                        if built is not None:
                            break
                if built is not None:
                    break
            if built is not None:
                break
        else: # try current dir too..?
            for f in os.listdir():
                for library_type in library_types:
                    if f.startswith(libname) and f.endswith(library_type):
                        ext = library_type
                        built = os.path.abspath(f)
                        target += ext
                        break
                if built is not None:
                    break
            else:
                target = None
                ext = None

        return built, target, ext

    def locate_lib(self, name=None, roots=None, extensions=None, library_types=(".so", ".pyd", ".dll")):
        """
        Tries to locate the build library file (if it exists)

        :return:
        :rtype:
        """

        if name is None:
            name = self.lib_name
        if roots is None:
            roots = [self.lib_dir]
        if os.path.isdir(name):
            roots = [os.path.dirname(name)] + roots
            name = os.path.basename(name)
        if extensions is None:
            extensions = ["", self.src_ext]

        built, target, ext = self.locate_library(name, roots, extensions, library_types=library_types)

        return self.locate_library(name, roots, extensions, library_types=library_types)

    def cleanup(self):
        # Locate the library and copy it out (if it exists)

        built, target, ext = self.locate_lib()

        if built is not None:
            target_dir = self.out_dir
            if target_dir is None:
                target_dir = self.lib_dir
            target = os.path.join(target_dir, target)

            if built != target:
                try:
                    os.remove(target)
                except:
                    pass
                os.rename(built, target)

            if self.cleanup_build:
                build_dir = os.path.join(self.src_dir, "build")
                if os.path.isdir:
                    shutil.rmtree(build_dir)

        return target


import json, numbers, os

from .Jobs import ExternalProgramJob, OptionsBlock, SystemBlock

__all__ = [
    "OrcaJob"
]

class OrcaOptionsBlock(OptionsBlock):
    opts_key = None

    job_params_json = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Templates', 'orca_opts.json')
    _json = None
    @classmethod
    def load_json(cls):
        if cls._json is None:
            with open(cls.job_params_json) as opts_json:
                cls._json = json.load(opts_json)
        return cls._json

    @classmethod
    def get_props(cls):
        return list(cls.load_json()[cls.opts_key])

class OrcaKeywordsBlock(OrcaOptionsBlock):
    opts_key = "Keywords"

    require_value = False

    def __init__(self, keywords=None, **rest):
        if keywords is not None:
            rest.update(keywords)
        super().__init__(**rest)

    _bs = None
    @classmethod
    def load_basis_sets(cls):
        if cls._bs is None:
            cls._bs = {k.lower():k for k in cls.load_json()['BasisSets']}
        return cls._bs

    @classmethod
    def get_props(cls):
        return ['keywords'] + super().get_props() + list(cls.load_basis_sets().values())

    def canonicalize_basis_set(self, k):
        bs = self.load_basis_sets()
        return bs.get(k.lower(), k)

    def get_params(self):
        kws = []
        for k,v in self.opts.items():
            if v is True:
                kws.append(k)
            else:
                if isinstance(v, (str, int, float, )):
                    kws.append(f'{k} {v}')
                else:
                    kws.append(k+" "+" ".join(str(v) for v in v))
        if len(kws) > 0:
            return {"keywords":"!" + " ".join(g for g in kws)}
        else:
            return {}

class OrcaGlobalsBlock(OrcaKeywordsBlock):
    opts_key = "Globals"
    require_value = True

    @classmethod
    def load_basis_sets(cls):
        return {}

    @classmethod
    def get_props(cls):
        return super().get_props()[1:]

    def get_params(self):
        kws = []
        for k,v in self.opts.items():
            if v is True:
                kws.append(k)
            else:
                if isinstance(v, (str, numbers.Number)):
                    kws.append(f'{k} {v}')
                else:
                    kws.append(k+" "+" ".join(v))
        if len(kws) > 0:
            return {"globals":"\n".join("%"+g for g in kws)}
        else:
            return {}

class OrcaMethodsBlock(OrcaOptionsBlock):
    opts_key = "Blocks"

    require_value = True

    def __init__(self, opts=None, **rest):
        if opts is not None:
            rest.update(opts)
        super().__init__(**rest)

    def format_options_block(self, header, opts):
        if opts is True:
            opts = {}
        if len(opts.items()) == 0:
            return ""
        else:
            padding = " " * len("%"+header)
            rows = "\n".join(padding + f"{k} {v}" for k,v in opts.items())
            return "\n".join(["%"+header, rows, padding+"end"])

    @classmethod
    def get_props(cls):
        return super().get_props() + ["opts"]

    def get_params(self):
        kws = []
        for k,v in self.opts.items():
            method_block = self.format_options_block(k, v)
            if len(method_block) > 0:
                kws.append(method_block)
        if len(kws) > 0:
            return {"methods":"\n".join(kws)}
        else:
            return {}

class OrcaSystemBlock(SystemBlock):
    __props__ = SystemBlock.__props__ + ("variables", "constants")

    fmt_key = ""

    def format_coordinate_block(self):
        charge = self.opts.get("charge", 0)
        multiplicity = self.opts.get("multiplicity", 1)
        carts = self.opts.get("cartesians")
        zmat = self.opts.get("zmatrix")

        coord_type = ("xyz" if carts is not None else "gzmt")

        chunks = []
        chunks.append(f"*{coord_type} {charge} {multiplicity}")

        if carts is not None:
            if isinstance(carts, str):
                chunks.append(carts)
            else:
                chunks.append(self.fmt_carts(self.opts.get('atoms'), carts))
        elif zmat is not None:
            if isinstance(zmat, str):
                chunks.append(zmat)
            else:
                chunks.append(self.fmt_zmat(self.opts.get('atoms'), zmat, self.opts.get('ordering')))
        else:
            raise ValueError("no coordinate spec supplied")

        chunks.append("*")

        return "\n".join(chunks)

    def get_params(self):
        base_opts = {}
        if len(self.opts) > 0:
            base_opts[self.fmt_key + "system"] = self.format_coordinate_block()
            if len(self.opts.get('bonds', [])) > 0:
                base_opts[self.fmt_key + "bonds"] = self.format_bonds_block()
        return base_opts

class OrcaJob(ExternalProgramJob):
    job_template = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Templates', 'orca_job.inp')
    blocks = [
        OrcaKeywordsBlock,
        OrcaGlobalsBlock,
        OrcaMethodsBlock,
        OrcaSystemBlock
    ]

    def __init__(self, *strs, basis_set=None, level_of_theory=None, **opts):
        if basis_set is not None:
            strs = (basis_set,) + strs
        if level_of_theory is not None:
            strs = (level_of_theory,) + strs
        # for o in strs:
        #     rt, o = OrcaKeywordsBlock.check_canon(o)
        #     if rt:
        #         opts[o] = True
        #     else:
        #         o[]
        opts = {o.lower():v for o,v in opts.items()}
        for o in strs:
            opts[o.lower()] = True
        super().__init__(**opts)

    # @classmethod
    # def get_extra_keys(cls):
    #     with open(cls.job_template) as r:
    #         t = r.read()
    #     return {s.strip("{").strip("}") for s in t.split()}

    @classmethod
    def get_block_types(cls):
        return cls.blocks

    @classmethod
    def load_template(cls):
        return cls.job_template

    # non_blank_line_terminated = {'link0', 'level_of_theory'}
    # def get_params(self):
    #     base_params = super().get_params()
    #     for k,b in base_params.items():
    #         if k not in self.non_blank_line_terminated:
    #             base_params[k] = b + "\n"
    #
    #     return base_params


"""
Provides an abstract base class off of which concrete surface implementations can be built
"""

import abc, numpy as np
from ... import Numputils as nput

__all__ = [
    "BaseSurface",
    "TaylorSeriesSurface",
    # "LinearExpansionSurface",
    # "LinearFitSurface",
    "InterpolatedSurface"
]

class BaseSurface(metaclass=abc.ABCMeta):
    """
    Surface base class which can be subclassed for relevant cases
    """
    def __init__(self, data, dimension):
        # we'll just give the core data a consistent attribute name
        self.data = data
        self.dimension = dimension
        # there's no 1000% general way to assess the dimension, so we just let it be passed
    @abc.abstractmethod
    def evaluate(self, points, **kwargs):
        """
        Evaluates the function at the points based off of "data"

        :param points:
        :type points:
        :return:
        :rtype:
        """
        raise NotImplemented
    def check_dimension(self, gridpoints, target=None, raise_exception=True):
        gridpoints = np.asanyarray(gridpoints)
        if target is None:
            target = self.dimension
        if target is not None:
            gp_dim = gridpoints.shape[-1]
            if gp_dim != target:
                if raise_exception:
                    raise ValueError(
                        "{}: dimension mismatch in call, grid points had dim {} but surface expects dim {}".format(
                            type(self).__name__,
                            gp_dim,
                            self.dimension
                        ))
                else:
                    return False
        return True
    def __call__(self, gridpoints, **kwargs):
        """

        :param gridpoints:
        :type gridpoints: np.ndarray
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        if nput.is_numeric(gridpoints):
            gridpoints = np.array([gridpoints])
        self.check_dimension(gridpoints)
        return self.evaluate(gridpoints, **kwargs)

    # def minimize(self, initial_guess, function_options=None, **opts):
    #     """
    #     Just calls into `scipy.optimize.minimize` as the default implementation
    #
    #     :param initial_guess: starting position for the minimzation
    #     :type initial_guess: np.ndarray
    #     :param function_options:
    #     :type function_options: dict | None
    #     :param opts:
    #     :type opts:
    #     :return:
    #     :rtype:
    #     """
    # #
    #     import scipy.optimize as opt
    #
    #     if function_options is None:
    #         function_options = {}
    #     test_call = self(initial_guess, **function_options) # if this fails the initial guess is bad
    #
    #     func = lambda x, f=self.evaluate, o=function_options: f(x, **o)
    #
    #     min = opt.minimize(func, initial_guess, **opts)
    #     return min.x




class TaylorSeriesSurface(BaseSurface):
    """
    A surface with an evaluator built off of a Taylor series expansion
    """
    def __init__(self, *derivs, dimension=None, **opts):
        """
        :param data: derivs or a tuple of derivs + options
        :type data:
        :param dimension:
        :type dimension:
        """
        from ..Taylor import FunctionExpansion

        if dimension is None:
            dimension = len(derivs[0])
        self.expansion = FunctionExpansion(derivs, **opts)
        opts["derivs"] = derivs
        super().__init__(opts, dimension)


    @property
    def center(self):
        return self.expansion.center
    @property
    def ref(self):
        return self.expansion.ref
    @property
    def expansion_tensors(self):
        return self.expansion.expansion_tensors

    def check_dimension(self, gridpoints, target=None, raise_exception=True):
        if target is not None or self.expansion.transforms is None:
            return super().check_dimension(gridpoints, target=target, raise_exception=raise_exception)
        else:
            d1, d2 = self.expansion.transforms[0][0].shape
            check_1 = super().check_dimension(gridpoints, target=d1, raise_exception=False)
            if check_1: return True
            check_2 = super().check_dimension(gridpoints, target=d2, raise_exception=False)
            if check_2: return True
            return super().check_dimension(gridpoints, target=target, raise_exception=raise_exception)


    def evaluate(self, points, **kwargs):
        """
        Since the Taylor expansion stuff is already built out this is super easy

        :param points:
        :type points:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return self.expansion(points, **kwargs)

class LinearExpansionSurface(BaseSurface):
    """
    A surface with an evaluator built off of an expansion in some user specified basis
    """
    def __init__(self, coefficients, basis=None, dimension=None):
        """
        :param coefficients: the expansion coefficients in the basis
        :type coefficients: np.ndarray
        :param basis: a basis of functions to use (defaults to power series)
        :type basis: Iterable[function] | None
        """
        if basis is None:
            basis = [(lambda x,n=n: x**n) for n in range(len(coefficients))]
        self.basis = basis
        self.coeffs = coefficients
        super().__init__({"basis":basis, "coeffs":coefficients}, dimension)

    def evaluate(self, points, **kwargs):
        """
        First we just apply the basis to the gridpoints, then we dot this into the coeffs

        :param points:
        :type points:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        basis_values = [f(points, **kwargs) for f in self.basis]
        return np.dot(self.coeffs, basis_values)

#TODO: add in a NonLinearFitSurface...
class LinearFitSurface(LinearExpansionSurface):
    """
    A surface built off of a LinearExpansionSurface, but done by fitting.
    The basis selection
    """
    def __init__(self,
                 points,
                 basis=None,
                 order=4,
                 dimension=None
                 ):
        """
        :param points: a set of points to fit to
        :type points: np.ndarray
        :param basis: a basis of functions to use (defaults to power series)
        :type basis: Iterable[function] | None
        """
        from ..FittableModels import LinearFittableModel, LinearFitBasis

        dim = points.shape[-1]
        if basis is None:
            basis = LinearFitBasis([LinearFitBasis.power_series]*dim, order=order)
        self.model = LinearFittableModel(basis)
        self.model.fit(points)
        self.fit_data = points
        super().__init__(self.model.parameters, basis, dimension)

    def evaluate(self, points, **kwargs):
        """

        :param points:
        :type points:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        return self.model.evaluate(points)

    # Is there any way to cleverly minimize a general function like this? My guess is I'd need to know how to take derivs
    def minimize(self, initial_guess=None, function_options=None, **opts):
        if initial_guess is None:
            initial_guess = self.fit_data[np.argmin(self.fit_data[:, -1])][:-1]
        return super().minimize(initial_guess, function_options=function_options, **opts)

class InterpolatedSurface(BaseSurface):
    """
    A surface that operates by doing an interpolation of passed mesh data
    """
    def __init__(self, xdata, ydata=None, dimension=None, **opts):
        from ..Interpolator import Interpolator
        if ydata is None:
            ydata = xdata[...,  -1]
            xdata = xdata[..., :-1]
        dimension = 1 if xdata.ndim == 1 else xdata.shape[-1]
        self.interp_data = (xdata, ydata)
        self.interp = Interpolator(xdata, ydata, **opts)
        super().__init__({"interpolator": self.interp}, dimension)

    def evaluate(self, points, **kwargs):
        """
        We delegate all the dirty work to the Interpolator so hopefully that's working...
        :param points:
        :type points:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return self.interp(points)

    def minimize(self, initial_guess=None, function_options=None, **opts):
        if initial_guess is None:
            initial_guess = self.interp_data[0][np.argmin(self.interp_data[1], axis=-1)]
        return super().minimize(initial_guess, function_options=function_options, **opts)



import uuid
import numpy as np

from .CoordinateSystem import CoordinateSystem
from .CoordinateSystemConverter import CoordinateSystemConverter
from ... import Numputils as nput
import weakref

__all__ = [
    "CompositeCoordinateSystem",
    "CompositeCoordinateSystemConverter",
    # "TransformedCoordinateSystem",
    # "TransformedCoordinateSystemConverter",
]

#TODO: these should all be metaclasses but :shrag:
class CompositeCoordinateSystem(CoordinateSystem):
    """
    Defines a coordinate system that comes from applying a transformation
    to another coordinate system
    """

    _register_cache = weakref.WeakValueDictionary()
    def __init__(self,
                 base_system,
                 conversion,
                 inverse_conversion=None,
                 jacobian=None,
                 inverse_jacobian=None,
                 name=None, batched=None, pointwise=True,
                 # max_expansion_order=0,
                 **opts):
        self.base_system = base_system
        self.conversion = conversion
        self.inverse_conversion = inverse_conversion
        self.pointwise = pointwise
        self.batched = batched if batched is not None else not pointwise
        # self.max_expansion_order = max_expansion_order
        super().__init__(**opts)
        self.name = self.canonical_name(name, conversion)
        self.fwd_jacobian = jacobian
        self.inverse_jacobian = inverse_jacobian
    @classmethod
    def canonical_name(cls, name, conversion):
        if name is None:
            if hasattr(conversion, 'name'):
                name = conversion.name
            else:
                name = str(uuid.uuid4()).replace("-", "")
        return name
        #
        #     return type('CompositeCoordinateSystem' + name, (cls,),
        #                 {'base_system': base_system, 'conversion': conversion, 'inverse': inverse})
        # # if self.base_system is None:
        # #     raise ValueError('{name} is a factory class and {name}.{method} should be used to register coordinate systems'.format(
        # #         name=type(self).__name__,
        # #         method='register'
        # #     ))
        # # super().__init__()

    @classmethod
    def register(cls, base_system, conversion, inverse_conversion=None, name=None, batched=None, pointwise=True, **opts):
        if (base_system, conversion) not in cls._register_cache:
            system_class = cls(base_system, conversion, inverse_conversion=inverse_conversion, name=name,
                               batched=batched, pointwise=pointwise, **opts)
            CompositeCoordinateSystemConverter(system_class).register()
            if system_class.inverse_conversion is not None:
                CompositeCoordinateSystemConverter(system_class, direction='inverse').register()
            cls._register_cache[(base_system, conversion)] = system_class
        return cls._register_cache[(base_system, conversion)]
    def unregister(self):
        raise NotImplementedError("destructor not here yet")
    def __repr__(self):
        return "{}({}, {})".format(
            type(self).__name__,
            self.base_system,
            self.name
        )


class CompositeCoordinateSystemConverter(CoordinateSystemConverter):
    def __init__(self, system:CompositeCoordinateSystem, direction='forward'):
        self.system = system
        self.direction = direction
    @property
    def types(self):
        if self.direction == 'inverse':
            return (self.system, self.system.base_system)
        else:
            return (self.system.base_system, self.system)
    def get_conversion(self):
        if self.direction == 'forward':
            convertser = self.system.conversion
        elif self.direction == 'inverse':
            convertser = self.system.inverse_conversion
        else:
            raise NotImplementedError("bad value for '{}': {}".format('direction', self.direction))
        return convertser
    def convert(self, coords, **kw):
        if self.system.pointwise:
            return nput.apply_by_coordinates(self.get_conversion(), coords, **kw)
        else:
            return self.get_conversion()(coords, **kw)
    def convert_many(self,
                     coords,
                     order=0,
                     derivs=None,
                     return_derivs=None,
                     **kw):
        # if self.system.max_expansion_order > 0:
        #     raise NotImplementedError(...)
        # else:
        if self.system.pointwise:
            base, opts = nput.apply_by_coordinates(self.get_conversion(), coords, **kw)
        elif self.system.batched:
            base, opts = self.convert(coords, **kw)
        else:
            base, opts = super().convert_many(coords, **kw)

        if return_derivs:
            if order is None or order == 0: order = 1

        if order > 0:
            if self.direction == 'inverse':
                test_jacobian = self.system.inverse_jacobian
                inv_jacobian = self.system.fwd_jacobian
            else:
                test_jacobian = self.system.fwd_jacobian
                inv_jacobian = self.system.inverse_jacobian

            if test_jacobian is not None:
                opts['derivs'] = test_jacobian(coords, order=order, **kw)
            elif inv_jacobian is not None:
                opts['derivs'] = nput.inverse_transformation(inv_jacobian(coords, order=order, **kw), order)

        return base, opts


# class TransformedCoordinateSystemConverter(CompositeCoordinateSystem):
#
#     def __init__(self, base_system, forward_transformation, reverse_transformation=None,
#                  name=None,
#                  batched=True, pointwise=False,
#                  **opts):
#         self.transformations = self.prep_transformations(forward_transformation, reverse_transformation)
#         super().__init__(base_system, self.apply_transformation, self.invert_transformation,
#                          name=name, batched=batched, pointwise=pointwise, **opts
#                          )
#
#     def prep_transformations(self):
#
#         if nput.is_numeric_array_like(forward_transformations):
#             forward_transformations = np.asanyarray(forward_transformations)
#             if forward_transformations.ndim == 2:
#                 forward_transformations = [forward_transformations]
#
#         if reverse_transformation is None:
#             reverse_transformation = nput.inverse_transformation(forward_transformations, len(forward_transformations))


from .CoordinateSystemConverter import CoordinateSystemConverter
from .CommonCoordinateSystems import CartesianCoordinates3D, SphericalCoordinates
from ...Numputils import vec_dots, vec_angles
import numpy as np
# this import gets bound at load time, so unfortunately PyCharm can't know just yet
# what properties its class will have and will try to claim that the files don't exist

class CartesianToSphericalConverter(CoordinateSystemConverter):
    """
    A converter class for going from Cartesian coordinates to ZMatrix coordinates
    """

    @property
    def types(self):
        return (CartesianCoordinates3D, SphericalCoordinates)

    def convert_many(self, coords, use_rad=True, origin=None, axes=None, **kw):
        """
        We'll implement this by having the ordering arg wrap around in coords?
        """
        coords = np.asanyarray(coords)
        if origin is None:
            origin = [0., 0., 0.]
        origin = np.asanyarray(origin)
        for d in range(coords.ndim - origin.ndim):
            origin = np.expand_dims(origin, 0)
        if axes is None:
            axes = [[1., 0., 0.], [0., 0., 1.]]
        axes = np.asanyarray(axes)
        for d in range(coords.ndim - axes[0].ndim):
            axes = [np.expand_dims(a, 0) for a in axes]

        coords = coords - origin
        dists = np.linalg.norm(coords, axis=-1)
        proj_axes = np.broadcast_to(axes[1], coords.shape)
        angles, _ = vec_angles(coords - vec_dots(proj_axes, coords)[..., np.newaxis]*proj_axes, axes[0]) # projected into x-y plane
        if not use_rad:
            angles = np.rad2deg(angles)
        polars, _ = vec_angles(coords, axes[1])
        if not use_rad:
            polars = np.rad2deg(polars)

        return (
            np.concatenate([np.expand_dims(x, -1) for x in [dists, angles, polars]], axis=-1),
            dict(**kw, origin=origin, axes=axes)
        )

    def convert(self, coords, ordering=None, use_rad=True, return_derivs=False, **kw):
        coords, opts = self.convert_many(np.expand_dims(coords[0]))
        coords = coords[0]
        opts['origin'] = opts['origin'][0]
        opts['axes'] = [a[0] for a in opts['axes']]
        return coords, opts

__converters__ = [ CartesianToSphericalConverter() ]


"""
Provides the conversion framework between coordinate systems
"""

from collections import OrderedDict as odict, deque
import os, abc, numpy as np, weakref
from ...Extensions import ModuleLoader
from ...Numputils import apply_by_coordinates

__all__ = [
    "CoordinateSystemConverters",
    "CoordinateSystemConverter",
    "SimpleCoordinateSystemConverter"
]

__reload_hook__ = ["...Extensions", '.CartesianToZMatrix', '.ZMatrixToCartesian']

######################################################################################################
##
##                                   CoordinateSystemConverter Class
##
######################################################################################################
class CoordinateSystemConverter(metaclass=abc.ABCMeta):
    """
    A base class for type converters
    """

    converters = None

    @property
    @abc.abstractmethod
    def types(self):
        """The types property of a converter returns the types the converter converts

        """
        pass

    def convert_many(self, coords_list, **kwargs):
        """Converts many coordinates. Used in cases where a CoordinateSet has higher dimension
        than its basis dimension. Should be overridden by a converted to provide efficient conversions
        where necessary.

        :param coords_list: many sets of coords
        :type coords_list: np.ndarray
        :param kwargs:
        :type kwargs:
        """
        return np.array([self.convert(coords, **kwargs) for coords in coords_list])

    @abc.abstractmethod
    def convert(self, coords, **kwargs):
        """The main necessary implementation method for a converter class.
        Provides the actual function that converts the coords set

        :param coords:
        :type coords: np.ndarray
        :param kwargs:
        :type kwargs:
        """
        pass

    def register(self, where=None, check=True):
        """
        Registers the CoordinateSystemConverter

        :return:
        :rtype:
        """
        if where is None:
            where = self.converters if not (self.converters, weakref.ref) else self.converters()
            if where is None:
                type(self).converters = weakref.ref(CoordinateSystemConverters)
                where = type(self).converters()
        where.register_converter(*self.types, self, check=check)

    def __call__(self, coords, **kwargs):
        if coords.ndim > 2: #TODO: make this a more robust check for the future
            return self.convert_many(coords, **kwargs)
        else:
            return self.convert(coords, **kwargs)

######################################################################################################
##
##                                   CoordinateSystemConverters Class
##
######################################################################################################
class CoordinateSystemConverters:
    """
    A coordinate converter class. It's a singleton so can't be instantiated.
    """

    converters = odict([])
    converter_graph = None
    converters_dir = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        "Resources",
        "Converters"
    )
    converters_package = ".".join(__name__.split(".")[:-1])
    converter_type = CoordinateSystemConverter

    def __init__(self):
        raise NotImplementedError("{} is a singleton".format(type(self)))

    @classmethod
    def get_coordinates(self, coordinate_set):
        """Extracts coordinates from a coordinate_set
        """
        pass

    @classmethod
    def _get_converter_file(self, file):
        if os.path.exists(file):
            abspath = file
        else:
            abspath = os.path.join(self.converters_dir, file)
        return abspath

    @classmethod
    def load_converter(self, converter):

        file = self._get_converter_file(converter)
        loader = ModuleLoader(self.converters_dir, self.converters_package)
        env = loader.load(file)

        try:
            converters = env.__converters__
        except KeyError:
            raise KeyError("converter at {} missing field '{}'".format(file, "converters"))

        for conv in converters:
            type_pair = tuple(conv.types)
            self.converters[type_pair] = conv

    _converters_loaded = False
    @classmethod
    def _preload_converters(self):
        """
        Preloads Cartesian/ZMatrix converters.
        Maybe will load others in the future.
        :return:
        :rtype:
        """

        if not self._converters_loaded:
            proxy_function = lambda x, y: x.is_compatible(x, y)
            self.converter_graph = ConversionGraph(proxy_function=proxy_function)

            from .DefaultConverters import __converters__ as converters
            for conv in converters:
                self._register(*conv.types, conv, move_to_end=True)

            if os.path.exists(self.converters_dir):
                for file in os.listdir(self.converters_dir):
                    if os.path.splitext(file)[1] == ".py":
                        self.load_converter(file)
            self._converters_loaded = True

    @classmethod
    def get_converter(cls, system1, system2):
        """
        Gets the appropriate converter for two CoordinateSystem objects

        :param system1:
        :type system1: CoordinateSystem
        :param system2:
        :type system2: CoordinateSystem
        :return:
        :rtype:
        """

        cls._preload_converters()
        if (system1, system2) in cls.converters:
            path = [(system1, system2)]
        else:
            path = cls.converter_graph.find_path_bfs(system1, system2)
        if path is None:
            raise KeyError(
                "{}: no rules for converting coordinate system {} to {} in {}".format(cls.__name__, system1, system2, [
                    "{}=>{}".format(
                        k.__name__ if isinstance(k, type) else type(k).__name__,
                        b.__name__ if isinstance(b, type) else type(b).__name__
                    ) for k, b in cls.converters
                ])
            )
        elif len(path) == 1:
            converter = cls.converters[path[0]]
        else:
            conversions = [(cls.converters[p], p) for p in path]
            return ChainedCoordinateSystemConverter([system1, system2], conversions)

        #
        # def _get_pathy_conversion(self, src, targ):
        #     conv_path = self._unit_graph.find_path_bfs(src, targ)
        #     if conv_path is None:
        #         return conv_path
        #     cval = 1
        #     for me, you in zip(conv_path, conv_path[1:]):
        #         cval *= self.data[(me, you)]["Value"]
        #     return cval
        #
        # try:
        #     converter = cls.converters[(system1, system2)]
        # except KeyError:
        #     for key_pair, conv in reversed(cls.converters.items()):
        #         if (
        #                 system1.has_conversion(system2)
        #                 or (system1.is_compatible(key_pair[0]) and system2.is_compatible(key_pair[1]))
        #         ):
        #             converter = conv
        #             break
        #     else:
        #         raise KeyError(
        #             "{}: no rules for converting coordinate system {} to {} in {}".format(cls.__name__, system1, system2, [
        #                 "{}=>{}".format(k.__name__, b.__name__ ) for k,b in cls.converters
        #             ])
        #         )

        return converter

    @classmethod
    def register_converter(cls, system1, system2, converter, check=True):
        """
        Registers a converter between two coordinate systems

        :param system1:
        :type system1: CoordinateSystem
        :param system2:
        :type system2: CoordinateSystem
        :return:
        :rtype:
        """
        cls._preload_converters()
        if check and not hasattr(converter, 'convert'): #isinstance(converter, cls.converter_type):
            raise TypeError('{}: registered converters should be subclasses of {} <{}> (got {} which inherits from {})'.format(
                cls.__name__,
                cls.converter_type, id(cls.converter_type),
                type(converter),
                ["{} <{}>".format(x, id(x)) for x in type(converter).__bases__]
            ))
        cls._register(system1, system2, converter)
    @classmethod
    def _register(cls, system1, system2, converter, move_to_end=False):
        cls.converters[(system1, system2)] = converter
        if move_to_end:
            cls.converters.move_to_end((system1, system2))
        graph = cls.converter_graph
        cls.converter_graph.add(system1, system2)
        for k in cls.converter_graph.keys():
            if graph.proxy_function(system1, k):
                cls.converters[(k, system2)] = converter
                if move_to_end:
                    cls.converters.move_to_end((k, system2))
            elif graph.proxy_function(system2, k):
                cls.converters[(system1, k)] = converter
                if move_to_end:
                    cls.converters.move_to_end((system1, k))

class ConversionGraph:
    """
    Pulled from the UnitGraph stuff
    """

    def __init__(self, stuff_to_update=(), proxy_function=None):
        self._graph = {}
        self.update(stuff_to_update)
        self.proxy_function = proxy_function

    def __contains__(self, item):
        return item in self._graph
    def add(self, node, connection):
        if node in self._graph:
            self._graph[node].add(connection)
        else:
            self._graph[node]={connection}
        if self.proxy_function is not None:
            for k in self._graph:
                if self.proxy_function(node, k):
                    self._graph[k].add(connection)
                elif self.proxy_function(connection, k):
                    self._graph[node].add(k)
        if not connection in self._graph:
            self._graph[connection] = set()
    def keys(self):
        return self._graph.keys()
    def update(self, iterable):
        for connection in iterable:
            self.add(*connection)
    def find_path_bfs(self, start, end):
        # we use a little poor-man's Dijkstra to find the shortest unit conversion path
        identity_function = self.proxy_function
        if identity_function is None:
            if not start in self._graph or not end in self._graph:
                return None
        else:
            if start not in self._graph:
                for k in self._graph:
                    if identity_function(start, k):
                        start = k
                        break
                else:
                    return None
            if end not in self._graph:
                for k in self._graph:
                    if identity_function(end, k):
                        end = k
                        break
                else:
                    return None

        q = deque() # deque as a FIFO queue
        q.append(start)
        parents = { start:None }
        steps = 0
        while len(q)>0:
            cur = q.pop()
            steps += 1
            for k in self._graph[cur]:
                if k is end:
                    parents[k] = cur
                    q = []
                    break
                elif k not in parents:
                    q.append(k)
                    parents[k] = cur
        if end not in parents:
            return None
        else:
            path = []*steps
            cur = end
            while cur is not start:
                nxt = parents[cur]
                path.append((nxt, cur))
                cur = nxt
            # path.append(start)
            return list(reversed(path))

class SimpleCoordinateSystemConverter(CoordinateSystemConverter):
    def __init__(self, types, conversion, **opts):
        super().__init__(**opts)
        self._types = types
        self.conversion = conversion
    @property
    def types(self):
        return self._types
    def convert(self, coords, **kw):
        return self.conversion(coords, **kw)
    def convert_many(self, coords, **kw):
        return self.convert(coords, **kw)
class ChainedCoordinateSystemConverter(CoordinateSystemConverter):

    def __init__(self, types, conversions, **opts):
        super().__init__(**opts)
        self._types = types
        self.conversions = conversions
    @property
    def types(self):
        return self._types
    def convert(self, crds, **kwargs):
        cur = crds
        for f, p in self.conversions:
            if hasattr(p[0], 'convert_coords') and not isinstance(p[0], type):
                cur = p[0].convert_coords(cur, p[1], converter=f, **kwargs)
            else:
                cur = f(cur, **kwargs)
            if isinstance(cur, tuple):
                cur, kwargs = cur
            else:
                kwargs = {}
        return cur, kwargs
    def convert_many(self, coords, **kw):
        return self.convert(coords, **kw)

CoordinateSystemConverter.converters = weakref.ref(CoordinateSystemConverters)



import pathlib
import json
import tempfile as tf

from . import Options as opts_handler

__all__ = [
    "is_filepath_like",
    "safe_open",
    "write_file",
    "read_file",
    "read_json",
    "write_json"
]

bad_file_chars = {" ", "\t", "\n"}
def is_filepath_like(file, bad_chars=None):
    if bad_chars is None:
        bad_chars = bad_file_chars
    return isinstance(file, pathlib.Path) or (
        isinstance(file, str)
        and all(b not in file for b in bad_chars)
    )

class safe_open:
    def __init__(self, file, **opts):
        self.file = file
        self._stream = None
        self.opts = opts
    def __enter__(self):
        if hasattr(self.file, 'seek'):
            return self.file
        else:
            self._stream = open(self.file, **self.opts)
            return self._stream.__enter__()
    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._stream is not None:
            return self._stream.__exit__(exc_type, exc_val, exc_tb)

class open_opts:
    mode: 'str'
    buffering: 'int'
    encoding: 'str | None'
    errors: 'str | None'
    newline: 'str | None'
    closefd: 'bool'
    opener: '(str, int)'

def write_file(file, data, mode='w+', **opts):
    with safe_open(file, mode=mode, **opts) as fs:
        fs.write(data)
    return file

def read_file(file, **opts):
    with safe_open(file, **opts) as fs:
        return fs.read()

def read_json(file, **opts):
    opts, js_opts = opts_handler.OptionsSet(opts).split(open_opts)
    with safe_open(file, **opts) as fs:
        return json.load(fs, **js_opts)

def write_json(file, data, **opts):
    opts, js_opts = opts_handler.OptionsSet(opts).split(open_opts)
    with safe_open(file, **opts) as fs:
        return json.dump(data, fs)


import os
import re
import json
from .. import Devutils as dev
from .JHTML import JHTML
from ..ExternalPrograms import PILInterface

__all__ = [
    "NotebookReader"
]

class NotebookReader:
    def __init__(self, json_or_fp):
        self._name = None
        self._nb_json = None
        self._init_data = json_or_fp

    class NotebookCell:
        def __init__(self, nb:'NotebookReader', cell_data):
            self.nb = nb
            self.data = cell_data
            self._attachments = None
        @property
        def cell_type(self):
            return self.data['cell_type']
        @property
        def cell_index(self):
            return self.data.get('cell_index', None)
        def __repr__(self):
            cls = type(self)
            return f"{cls.__name__}(<{self.cell_index}>, '{self.cell_type}')"
        @property
        def attachments(self):
            return self.prep_attachments(self.nb, self.data.get('attachments', {}))
        def get_images(self):
            return {
                k:PILInterface.from_url(v.attrs['src'])
                for k,v in self.attachments.items()
                if k.startswith('image/')
            }
        @property
        def source(self):
            return self.data["source"]
        @property
        def text(self):
            if isinstance(self.data['source'], str):
                return self.data['source']
            else:
                return ''.join(self.data['source'])

        @classmethod
        def get_cell_header(cls, data):
            src = data['source']
            if len(src) == 0: return None

            if not isinstance(src, str):
                src = src[0]
            if src.strip().startswith("#"):
                return src.splitlines()[0].strip("#").strip()
            else:
                return None
        @property
        def cell_header(self):
            return self.get_cell_header(self.data)
        @property
        def html(self):
            return self.nb.get_mime_type_loader("text/html")(self.text)
        @classmethod
        def prep_attachments(cls, nb, attachment_data):
            attachments = {}
            for file_name,attachment_data in attachment_data.items():
                return {
                    mime_type: nb.get_mime_type_loader(mime_type)(value)
                    for mime_type, value in attachment_data.items()
                }
            return attachments

    def cell(self, data):
        return self.NotebookCell(self, data)
    def get_mime_image_loader(self, img_type):
        def load(data):
            return JHTML.Image(src=f"data:{img_type};base64,{data}")
        return load
    def get_mime_type_loader(self, mime_type):
        if mime_type.split("/")[0] == "image":
            return self.get_mime_image_loader(mime_type)
        else:
            return JHTML.Pre

    class CellList:
        def __init__(self, nb:'NotebookReader', cell_list):
            self.nb = nb
            self.cell_list = cell_list

        def __repr__(self):
            cls = type(self)
            return f"{cls.__name__}(<{len(self.cell_list)}>)"
        def __len__(self):
            return len(self.cell_list)
        def __iter__(self):
            for c in self.cell_list:
                yield self.nb.cell(c)
        def __getitem__(self, item):
            return self.nb.cell(self.cell_list[item])

        def modify(self, cells=None, *, nb=None):
            if nb is None:
                nb = self.nb
            return type(self)(nb, cells)
        def filter_cells(self, filter):
            return self.modify(
                [
                    dict(c, cell_index=i)
                        if 'cell_index' not in c
                    else c
                    for i, c in enumerate(self.cell_list)
                    if filter(c)
                ]
            )
        def find_cells_by_type(self, pattern):
            pattern = self.prep_regex(pattern)
            return self.filter_cells(
                lambda c: re.match(pattern, c['cell_type'])
            )
        def _match_header(self, pattern, data):
            header = self.nb.NotebookCell.get_cell_header(data)
            if header is None:
                return False
            else:
                return re.match(pattern, header)
        def find_cells_by_header(self, pattern):
            pattern = self.prep_regex(pattern)
            return self.filter_cells(
                lambda c: self._match_header(pattern, c)
            )
        def find_cells_with_attachments(self, pattern=None):
            if pattern is None:
                pattern = ".+"
            pattern = self.prep_regex(pattern)
            return self.filter_cells(
                lambda c: any(
                    any(re.match(pattern, mime) for mime in a_dat.keys())
                    for a_dat in c.get('attachments', {}).values()
                )
            )

        @classmethod
        def prep_regex(self, pattern, flags=None):
            if isinstance(pattern, tuple):
                pattern = "|".join(pattern)
            if isinstance(pattern, str):
                if flags is None:
                    pattern = re.compile(pattern)
                else:
                    pattern = re.compile(pattern, flags=flags)
            return pattern

    def cell_list(self, cells=None):
        if cells is None:
            cells = self.nb_json['cells']
        if isinstance(cells, dict):
            cells = [cells]
        return self.CellList(self, cells)

    def __repr__(self):
        cls = type(self)
        return f"{cls.__name__}('{self.file_name}')"

    def load_notebook(self, nb_js):
        if isinstance(nb_js, dict):
            name = self.get_notebook_name(nb_js)
        elif isinstance(nb_js, str) and nb_js.strip().startswith('{'):
            name = self.get_notebook_name(nb_js)
            nb_js = json.loads(nb_js)
        else:
            name = nb_js
            nb_js = dev.read_json(nb_js)
        return name, nb_js
    def get_notebook_name(self, nb_js=None):
        if nb_js is None:
            nb_js = self.nb_json
        md_cells = self.cell_list(nb_js['cells']).find_cells_by_type('markdown')
        if len(md_cells) > 0:
            for c in md_cells:
                h = c.cell_header
                if h is not None and len(h) > 0: return h
            else:
                return md_cells[0].text.splitlines()[0].strip()
        else:
            return "Unnamed"
    @property
    def file_name(self):
        if self._nb_json is None:
            self._name, self._nb_json = self.load_notebook(self._init_data)
        return self._name

    @property
    def nb_json(self):
        if self._nb_json is None:
            self._name, self._nb_json = self.load_notebook(self._init_data)
        return self._nb_json


    @classmethod
    def get_notebook_files(cls, directory='.'):
        return [
            o for o in os.listdir(directory)
            if os.path.isfile(o) and os.path.splitext(o)[-1] == '.ipynb'
        ]
    @classmethod
    def sort_by_evaluation_time(cls, file_list, directory='.'):
        return list(sorted(
            file_list,
            key=lambda o:os.path.getmtime(os.path.join(directory, o))
        ))
    @classmethod
    def active_notebook(cls, directory='.'):
        nb_files = cls.sort_by_evaluation_time(cls.get_notebook_files(directory))
        if len(nb_files) == 0:
            return None
        else:
            return cls(nb_files[0])



"""
Provides a set of singleton objects that can declare their purpose a little bit better than None can
"""
import itertools
# import enum
import types
import numbers

__all__ = [
    "default",
    "is_default",
    "handle_default",
    "uninitialized",
    "is_uninitialized",
    "handle_uninitialized",
    "missing",
    "is_missing",
    "is_interface_like",
    "is_dict_like",
    "is_option_spec_like",
    "destructure_option_spec",
    "is_list_like",
    "is_number",
    "is_int",
    "is_atomic",
    "cached_eval",
    "merge_dicts",
    "str_comp",
    "str_is",
    "str_in",
    "str_elide",
    "resolve_key_collision",
    "merge_dicts",
    "context_wrap",
    "slice_dict",
    "dict_take"
]

class SingletonType:
    """
    A base type for singletons
    """
    __slots__ = []
    def __eq__(self, other):
        return self is other or (
            type(self).__name__ == type(other).__name__
            and self.__module__ == other.__module__
        )

class DefaultType(SingletonType):
    """
    A type for declaring an argument should use its default value (for when `None` has meaning)
    """
    __is_default__ = True
default=DefaultType()

class MissingType(SingletonType):
    """
    A type for declaring a value is missing (for when `None` has meaning)
    """
    __is_missing__ = False
missing=MissingType()

class UninitializedType(SingletonType):
    """
    A type for declaring an argument should use its default value (for when `None` has meaning)
    """
    __is_uninitialized__ = True
uninitialized = UninitializedType()

def is_atomic(obj,
              interface_types=(str, bool, numbers.Number),
              exlusion_types=None,
              implementation_props=None
              ):
    return is_interface_like(obj, interface_types, exlusion_types, implementation_props)

def is_number(obj,
              interface_types=(numbers.Number,),
              exlusion_types=None,
              implementation_props=None
              ):
    return is_interface_like(obj, interface_types, exlusion_types, implementation_props)

def is_int(obj,
           interface_types=(numbers.Integral,),
           exlusion_types=None,
           implementation_props=None
           ):
    return is_interface_like(obj, interface_types, exlusion_types, implementation_props)

def is_interface_like(obj, interface_types, exlusion_types, implementation_attrs):
    return (
            (exlusion_types is None)
            or not isinstance(obj, exlusion_types)
    ) and (
            (interface_types is not None and isinstance(obj, interface_types))
            or (
                implementation_attrs is not None
                and all(hasattr(obj, a) for a in implementation_attrs)
            )
    )

def is_dict_like(obj,
                 interface_types=(dict, types.MappingProxyType),
                 exlusion_types=None,
                 implementation_props=('items',)
                 ):
    return is_interface_like(obj, interface_types, exlusion_types, implementation_props)

def is_list_like(obj,
                 interface_types=(list, tuple),
                 exlusion_types=(str, dict, type),
                 implementation_props=('__getitem__',)
                 ):
    return is_interface_like(obj, interface_types, exlusion_types, implementation_props)

def is_default(obj, allow_None=True):
    if allow_None and obj is None:
        return True

    return (
            obj is default
            or isinstance(obj, DefaultType)
            or (hasattr(obj, '__is_default__') and obj.__is_default__)
    )

def is_option_spec_like(obj, allow_enums=True):
    method, opts = destructure_option_spec(obj, allow_enums)
    return method is not None

def destructure_option_spec(spec, allow_enums=True, method_key='method'):
    if isinstance(spec, (str, bool)) or is_number(spec) or is_default(spec):
        opts = {}
        method = spec
    elif (
            allow_enums and
            hasattr(spec, 'name') and hasattr(spec, 'value')
    ):  # enum
        method = spec.value
        opts = {}
    elif is_dict_like(spec):
        opts = spec.copy()
        method = opts.pop(method_key, None)
    elif callable(spec):
        method = spec
        opts = {}
    else:
        try:
            method, opts = spec
        except TypeError:
            method = None
            opts = None

    return method, opts

def handle_default(opt, default_value, allow_None=True):
    if is_default(opt, allow_None=allow_None):
        return default_value
    else:
        return opt

def is_uninitialized(obj, allow_None=True):
    if allow_None and obj is None:
        return True

    return (
            obj is uninitialized
            or isinstance(obj, UninitializedType)
            or (hasattr(obj, '__is_uninitialized__') and obj.__is_uninitialized__)
    )

def handle_uninitialized(opt, initializer, allow_None=True, args=(), kwargs=None):
    if is_uninitialized(opt, allow_None=allow_None):
        return initializer(*args, **({} if kwargs is None else kwargs))
    else:
        return opt

def is_missing(obj, allow_None=True):
    if allow_None and obj is None:
        return True

    return (
            obj is missing
            or isinstance(obj, MissingType)
            or (hasattr(obj, '__is_missing__') and obj.__is_missing__)
    )


def cached_eval(cache, key, generator, *,
                condition=None,
                args=(),
                kwargs=None):
    condition = (condition is None or condition(key))
    if not condition:
        if kwargs is None: kwargs = {}
        return generator(*args, **kwargs)

    if key in cache:
        return cache[key]

    if kwargs is None: kwargs = {}
    val = generator(*args, **kwargs)
    cache[key] = val

    return val

def str_comp(str_val, test, test_val):
    return isinstance(str_val, str) and test(str_val, test_val)
def str_is(str_val, test_val):
    return isinstance(str_val, str) and str_val == test_val
def str_in(str_val, test_vals):
    return isinstance(str_val, str) and str_val in test_vals
def str_elide(long_str, width=80, placeholder='...'):
    l = len(long_str)
    if l > width:
        total_width = width - len(placeholder)
        l = total_width // 2 + (total_width % 2)
        r = total_width // 2
        long_str = long_str[:l] + placeholder + long_str[-r:]
    return long_str

def resolve_key_collision(a, b, k, merge_iterables=True):
    if is_dict_like(a[k]):
        if not is_dict_like(b[k]):
            return b[k]
            # raise ValueError(f"can't resolve key collision on key {k} between {a[k]} and {b[k]}")
        return merge_dicts(a[k], b[k], resolve_key_collision, merge_iterables=merge_iterables)
    elif merge_iterables and isinstance(a[k], set):
        if not isinstance(b[k], set):
            return b[k]
        a = set(a[k])
        a.update(b)
        return a
    elif merge_iterables and is_list_like(a[k]):
        if not is_list_like(b[k]):
            return type(a[k])(
                itertools.chain(a[k], b[k])
            )
        else:
            return type(a[k])(
                itertools.chain(a[k], [b[k]])
            )
    elif merge_iterables and is_list_like(b[k]):
        return type(b[k])(
            itertools.chain([a[k]], b[k])
        )
    else:
        return b[k]

def merge_dicts(a, b, collision_handler=None, merge_iterables=True):
    key_inter = a.keys() & b.keys()
    diff_a = a.keys() - key_inter
    diff_b = b.keys() - key_inter
    dd = {k: a[k] for k in diff_a}
    dd.update((k, b[k]) for k in diff_b)
    if len(key_inter) > 0:
        if collision_handler is None:
            collision_handler = lambda a, b, k:resolve_key_collision(a, b, k, merge_iterables=merge_iterables)
        dd.update(
            (k, collision_handler(a, b, k))
            for k in key_inter
        )

    return dd

class context_wrap:
    def __init__(self, obj):
        self.obj = obj
    def __enter__(self):
        if hasattr(self.obj, '__enter__'):
            return self.obj.__enter__()
        else:
            return self.obj
    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self.obj, '__exit__'):
            return self.obj.__exit__(exc_type, exc_val, exc_tb)

class slice_dict:
    __slots__ = ["dict_obj"]
    def __init__(self, dict_obj:types.MappingProxyType):
        self.dict_obj = dict_obj
    def __getitem__(self, item):
        return dict_take(self.dict_obj, item)

def dict_take(dict_obj:types.MappingProxyType, spec):
    if is_number(spec):
        if spec < 0:
            spec = len(dict_obj) - spec
        for k in itertools.islice(dict_obj.keys(), spec, -1):
            return k, dict_obj[k]
    elif isinstance(spec, slice):
        return dict(
            itertools.islice(dict_obj.items(), spec.start, spec.stop, spec.step)
        )
    else:
        if not is_int(spec[0]):
            return {
                k:dict_obj[k]
                for k in spec
            }

        results = {}
        rem = set(spec)
        for n,k in enumerate(dict_obj.keys()):
            if n in rem:
                results[n] = (k, dict_obj[k])
                rem.remove(n)
                if len(rem) == 0: break
        else:
            raise IndexError("can't take elements {rem}")

        return dict(results[n] for n in spec)


import typing
from .Parallelizers import Parallelizer

__all__ = [
    "ClientServerRunner"
]

class ClientServerRunner:
    """
    Provides a framework for running MPI-like scripts in a client/server
    model
    """

    def __init__(self, client_runner:typing.Callable, server_runner:typing.Callable, parallelizer:Parallelizer):
        self.client = client_runner
        self.server = server_runner
        self.par = parallelizer

    def run(self):
        """
        Runs the client/server processes depending on if the parallelizer
        is on the main or server processes

        :return:
        :rtype:
        """
        if self.par.on_main:
            self.client()
        else:
            self.server()






import numpy as np
import itertools

from .. import Numputils as nput
from .. import Iterators as itut
from .Permutations import UniquePermutations, UniqueSubsets, IntegerPartitioner, IntegerPartitioner2D, lehmer_encode

__all__ = [
    "YoungTableauxGenerator"
]

class YoungTableauxGenerator:
    #TODO: save these to data file for future use

    tableaxu_cache = {}
    subset_cache = {}
    def __init__(self, base_int, tableaxu_cache=None, subset_cache=None):
        self.base = base_int
        if tableaxu_cache is None:
            tableaxu_cache = self.tableaxu_cache
        self.cache = tableaxu_cache
        if subset_cache is None:
            subset_cache = self.subset_cache
        self.subset_cache = subset_cache

    def number_of_tableaux(self, partitions=None, **partition_opts):
        if partitions is None:
            partitions = IntegerPartitioner.partitions(self.base, **partition_opts)

        smol = nput.is_numeric(partitions[0])
        if smol:
            partitions = [partitions]

        if not nput.is_numeric(partitions[0][0]):
            partitions = itertools.chain(*partitions)
        else:
            partitions = partitions
        return sum(self.count_standard_tableaux(p) for p in partitions)
    def get_standard_tableaux(self, partitions=None, *, symbols=None,  brute_force=False,
                              return_partitions=False,
                              **partition_opts):
        if partitions is None:
            partitions = IntegerPartitioner.partitions(self.base, **partition_opts)

        smol = nput.is_numeric(partitions[0])
        if smol:
            partitions = [partitions]

        if not nput.is_numeric(partitions[0][0]):
            part_iter = itertools.chain(*partitions)
            if return_partitions:
                partitions = list(part_iter)
                part_iter = partitions
        else:
            part_iter = partitions

        tabs = [
            self.standard_partition_tableaux(partition,
                                             cache=self.cache,
                                             subset_cache=self.subset_cache,
                                             brute_force=brute_force, symbols=symbols)
            for partition in part_iter
        ]
        if smol:
            tabs = tabs[0]

        if return_partitions:
            if smol:
                partitions = partitions[0]
            return tabs, partitions
        else:
            return tabs

    @classmethod
    def standard_partition_tableaux_bf(cls, partition, unique_perms=False, concatenate=False):
        if unique_perms:
            n = len(partition)
            perm_list = sum(([n - i] * k for i, k in enumerate(partition)), [])
            #
            perms = UniquePermutations(perm_list[1:]).permutations()
            perms = np.concatenate([
                np.full((len(perms), 1), n, dtype=perms.dtype),
                perms
            ], axis=1)
            idx_pos = np.argsort(np.argsort(-perms, axis=1))
        else:
            idx_pos = nput.permutation_indices(sum(partition), sum(partition))

        tableaux = np.array_split(idx_pos, np.cumsum(partition)[:-1], axis=1)
        valid = np.full(len(idx_pos), True)
        for i, t in enumerate(zip(*tableaux)):
            if any(len(tt) > 1 and (np.diff(tt) < 0).any() for tt in t):
                valid[i] = False
            if valid[i] and any(len(tt) > 1 and (np.diff(tt) < 0).any() for tt in itut.transpose(t)):
                valid[i] = False

        tableaux = [t[valid] for t in tableaux]
        if concatenate:
            tableaux = np.concatenate(tableaux, axis=1)

        # if return_perms:
        #     return perms[valid], tableaux
        # else:
        return tableaux

    @classmethod
    def populate_sst_frames(cls, partition, frame, segment_lists):
        frame_list = []
        for segments in segment_lists:
            splits = [0] * len(partition)
            subframe = [np.zeros(k, dtype=int) for k in partition]
            for s, r in zip(subframe, frame):
                k = 0
                for n, i in enumerate(r):
                    j = splits[n]
                    s[k:k + i] = segments[n][j:j + i]
                    splits[n] += i
                    k += i
            frame_list.append(subframe)
        return frame_list

    @classmethod
    def standard_partitions(cls, partition):
        #TODO: make this more efficient
        base_partitions = IntegerPartitioner2D.get_partitions(partition, partition)
        offsets = np.cumsum(base_partitions, axis=-1)
        valid = np.all(np.all(np.diff(offsets, axis=-2) <= 0, axis=-2), axis=-1)

        return [
            base_partitions[valid,],
            offsets[valid,],
        ]

    @classmethod
    def hook_numbers(cls, partition):
        return [
            (p - j) + sum(
                1 if pk >= (j+1) else 0
                for pk in partition[i + 1:]
            )
            for i, p in enumerate(partition)
            for j in range(p)
        ]

    @classmethod
    def count_standard_tableaux(cls, partition):
        nums = np.arange(1, np.sum(partition)+1)
        denoms = np.sort(cls.hook_numbers(partition))
        return np.round(np.prod(nums/denoms)).astype(int)

    @classmethod
    def split_frame(cls, partition, offsets):
        # `offsets` is just a vector of offset indices
        # we find the ways we need to split the partition
        frame_list = []
        mp = np.max(partition)
        partition = np.array(partition)
        offsets = np.array(offsets)
        for i in range(len(partition)):
            p = partition[i]
            if p == 0: continue
            o = offsets[i]
            frame_list.append([
                np.pad([p], [i, len(partition)-(i+1)]),
                np.pad([o], [i, len(partition)-(i+1)])
            ])
            # diffs = np.zeros_like(offsets)
            # diffs[i + 1:] = o - offsets[i + 1:]
            # col_offset = np.concatenate([np.full(i+1, o), offsets[i+1:]])
            # this_part = np.clip(partition - diffs, 0, mp)
            # frame_list.append([this_part, col_offset])
            # partition = np.min([partition, diffs], axis=0)
            # offsets[i+1:] += partition[i+1:]
        return frame_list

    @classmethod
    def _sst_2(cls, partition, cache=None, subset_cache=None, symbols=None):
        tableaux = None
        if len(partition) == 1:
            tableaux = [
                np.arange(partition[0], dtype=int)[np.newaxis]
            ]
        elif sum(partition) == 2:
            tableaux = [
                np.array([[0]], dtype=int),
                np.array([[1]], dtype=int)
            ]
        elif sum(partition) == 3:
            if partition[0] == 2:
                tableaux = [
                    np.array([[0, 1], [0, 2]], dtype=int),
                    np.array([[2], [1]], dtype=int)
                ]
            else:
                tableaux = [
                    np.array([[0]], dtype=int),
                    np.array([[1]], dtype=int),
                    np.array([[2]], dtype=int),
                ]
        if tableaux is not None:
            if symbols is not None:
                symbols = np.asanyarray(symbols)
                return [
                    symbols[tab]
                    for tab in tableaux
                ]

            return tableaux

        if cache is None:
            cache = {}
        if subset_cache is None:
            subset_cache = {}
        partition = tuple(partition)
        if partition in cache:
            tableaux = cache[partition]
            if symbols is not None:
                symbols = np.asanyarray(symbols)
                return [
                    symbols[tab]
                    for tab in tableaux
                ]

            return tableaux
        else:
            frames, offsets = cls.standard_partitions(partition)
            offsets[:, :, 1:] = offsets[:, :, :-1]
            offsets[:, :, 0] = 0

            if symbols is None:
                symbols = np.arange(np.sum(partition))
            segments = np.array_split(symbols, np.cumsum(partition)[:-1])

            # n_frames = 0
            tableaux_generators = []
            for f, o in zip(frames, offsets):
                # print("!", f, o)
                # nsubs = 1
                subframes = []
                for pp, oo, seg in zip(f.T, o.T, segments):
                    # print("|||", pp, oo, seg)
                    frame_splits = cls.split_frame(pp, oo)
                    # for p,o in frame_splits:
                    #     print(">", p)
                    partition_sizes = [
                        np.sum(p) for p, o in frame_splits
                    ]

                    tp_key = tuple(partition_sizes)
                    subsets = subset_cache.get(tp_key)
                    if subsets is None:
                        subsets = UniqueSubsets.unique_subsets(partition_sizes)
                        subset_cache[tp_key] = subsets

                    comb_splits = np.array_split(
                        subsets,
                        np.cumsum(partition_sizes)[:-1],
                        axis=1
                    )
                    # print(">><<>><<", partition_sizes)
                    # print(subsets)
                    subssts = []
                    for spl in zip(*comb_splits):
                        subrow = []
                        # print('...', spl)
                        for (p, o), ss in zip(frame_splits, spl):
                            if len(ss) == 0: continue
                            # print(":", seg[ss])
                            p_idx = [x for x, p0 in enumerate(p) if p0 > 0]
                            subsubssts = cls._sst_2([p[x] for x in p_idx], cache=cache, symbols=seg[ss])
                            reshaped_ssts = []
                            for ss in zip(*subsubssts):
                                p_full = [[]] * len(p)
                                for x, r in zip(p_idx, ss):
                                    p_full[x] = r
                                reshaped_ssts.append([o, p_full])
                            subrow.append(reshaped_ssts)
                        # subrow_joins = []
                        # for r in itertools.product(*subrow):
                        #     joins = [
                        #         np.concatenate([rr[1][i] for rr in r])
                        #         for i in range(len(r[0]))
                        #     ]
                        #     print(joins)
                        #     if all(np.all(j == np.sort(j)) for j in joins):
                        #         subrow_joins.append(r)
                        # subssts.append(subrow_joins)#
                        subssts.append(list(itertools.product(*subrow)))
                        # nsubs *= len(subssts[-1])

                    subframes.append(subssts)
                # n_frames += nsubs
                tableaux_generators.append(subframes)

            n_frames = sum(
                np.prod([
                    len(f)
                    for f in frame_list
                ])
                for subframes in tableaux_generators
                for frame_list in itertools.product(*subframes)
            )

            mask = np.full(n_frames, True)
            tableaux = [np.full((n_frames, p), -1, dtype=int) for p in partition]
            n = 0
            for subframes in tableaux_generators:
                for frame_list in itertools.product(*subframes):
                    for frame_choice in itertools.product(*frame_list):
                        for f in frame_choice:
                            for oo, p_sets in f:
                                if not mask[n]: break
                                for i, (o, ss) in enumerate(zip(oo, p_sets)):
                                    k = len(ss)
                                    if k > 0:
                                        for j in range(i):
                                            if np.any(tableaux[j][n, o:o + k] >= ss):
                                                mask[n] = False
                                                break
                                        if not mask[n]: break
                                        tableaux[i][n, o:o + k] = ss
                        n += 1

            tableaux = [
                t[mask, :]
                for t in tableaux
            ]

            mask = np.all([
                np.all(t > -1, axis=1)
                for t in tableaux
            ], axis=0)
            tableaux = [
                t[mask, :]
                for t in tableaux
            ]


            codes = lehmer_encode(np.concatenate(tableaux, axis=-1))
            _, mask = np.unique(codes, return_index=True)
            tableaux = [
                t[mask, :]
                for t in tableaux
            ]

            cache[partition] = tableaux

            return tableaux
    @classmethod
    def standard_partition_tableaux(cls, partition,
                                    cache=None,
                                    subset_cache=None,
                                    symbols=None, brute_force=False):
        if brute_force:
            tableaux = cls.standard_partition_tableaux_bf(partition, concatenate=False, unique_perms=False)
        else:
            tableaux = cls._sst_2(partition, cache=cache, subset_cache=subset_cache)
        if symbols is not None:
            symbols = np.asanyarray(symbols)
            tableaux = [
                symbols[tab]
                for tab in tableaux
            ]
        return tableaux

    @classmethod
    def print_tableaux(cls, tableaux):
        from ..Formatters import TableFormatter
        if isinstance(tableaux[0], np.ndarray):
            tableaux = [tableaux]
        for tabs in tableaux:
            for sst in zip(*tabs):
                tab = TableFormatter("").format(sst)
                print("-"*len(tab.split("\n")[0]))
                print(tab)



import numpy as np
from .TableFormatters import TableFormatter

__all__ = [
    "format_tensor_element_table",
    "format_symmetric_tensor_elements",
    "format_mode_labels",
    "format_zmatrix"
]

def format_tensor_element_table(inds, vals,
                                headers=("Indices", "Value"),
                                format="{:8.3f}",
                                column_join="|",
                                index_format="{:>5.0f}",
                                **etc
                                ):
    if isinstance(column_join, str):
        column_join = (" ", column_join)
    vals = np.asanyarray(vals)
    if vals.ndim == 1:
        vals = vals[:, np.newaxis]
    spans = [len(inds), vals.shape[-1]]
    return TableFormatter(
        column_formats=[index_format] * spans[0] + (
            [format] * spans[1]
                if isinstance(format, str) else
            format
        ),
        headers=headers,
        header_spans=spans,
        column_join=(
            [column_join[0]] * (spans[0]-1)
            + [column_join[1]]
            + [column_join[0]] * (vals.shape[-1]-1)
        ),
        **etc
    ).format(np.concatenate([np.array(inds).T, vals], axis=1))

def format_symmetric_tensor_elements(
        tensor,
        symmetries=None,
        cutoff=1e-6,
        headers=("Indices", "Value"),
        format="{:12.3f}",
        **etc
):
    tensor = np.asanyarray(tensor)
    if symmetries is None:
        symmetries = [np.arange(tensor.ndim)]

    symmetries = [np.sort(s) for s in symmetries]

    inds = np.where(np.abs(tensor) >= cutoff)
    if len(symmetries) > 0:
        inds_tests = [
            np.all(np.diff([inds[i] for i in symm], axis=0) >= 0, axis=0)
            for symm in symmetries
        ]
        inds_mask = np.all(inds_tests, axis=0)
        inds = tuple(x[inds_mask] for x in inds)

    vals = tensor[inds]

    return format_tensor_element_table(inds, vals, headers=headers, format=format, **etc)

def format_mode_labels(labels,
                       freqs=None,
                       high_to_low=True,
                       mode_index_format="{:.0f}",
                       frequency_format="{:.0f}",
                       headers=None,
                       column_join=" | ",
                       none_tag="mixed",
                       **etc
                       ):
    labels = [
        none_tag
            if lab.type is None else
        " ".join(t for t in lab.type if t is not None)
        for lab in labels
    ]

    if freqs is not None:
        return TableFormatter(
            [mode_index_format, frequency_format, "{}"],
            headers=("Mode", "Frequency", "Label") if headers is None else headers,
            column_join=column_join,
            **etc
        ).format([
            [i+1, f, lab]
            for i,(f,lab) in enumerate(zip(
                reversed(freqs) if high_to_low else freqs,
                reversed(labels) if high_to_low else labels
            ))
        ])
    else:
        return TableFormatter(
            [mode_index_format, "{}"],
            headers=("Mode", "Label") if headers is None else headers,
            column_join=column_join,
            **etc
        ).format([
            [i + 1, lab]
            for i, lab in enumerate(
                reversed(labels) if high_to_low else labels
            )
        ])

def format_zmatrix(zm, preserve_embedding=True, preserve_indices=True, list_form=True):
    max_ind = int(np.log10(max([max(z) for z in zm]))) + 2
    formatter = TableFormatter(
        f"{{:>{max_ind}.0f}}",
        column_join=", " if list_form else " ",
        row_padding=" [" if list_form else "",
        row_join="],\n" if list_form else "\n"
    )
    if not preserve_embedding:
        if isinstance(zm, np.ndarray):
            zm = zm.tolist()
        else:
            zm = [list(z) for z in zm]
        if len(zm[0]) == 4:
            zm[0][1] = ""
            zm[0][2] = ""
            zm[0][3] = ""

            zm[1][2] = ""
            zm[1][3] = ""

            zm[2][3] = ""
        else:
            zm[0][1] = ""
            zm[0][2] = ""

            zm[1][2] = ""

    if not preserve_indices:
        if len(zm[0]) == 4:
            zm = [z[1:] for z in zm[1:]]

    base = formatter.format(zm)
    if list_form:
        base = "[" + base[1:] + "]]"

    return base
